#!/usr/bin/env python
#
# Copyright (C) 2017 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Outputs the warnings that are common to all builders.

Suppressed tests that are nonetheless passing are output as warnings
by vogar.  Any tests that generate warnings in every builder are good
candidates for no longer being suppressed, since they're passing on
a regular basis."""

import collections
import json
import requests

# The number of recent builds to check for each builder
NUM_BUILDS = 5
# The buildbot step to check for warnings
BUILDBOT_STEP = 'test libcore'


def main():
    # Dict from builder+build_num combination to the list of warnings
    # in that build
    warnings = collections.defaultdict(list)
    r = requests.get('https://build.chromium.org/p/client.art/json/builders')
    if r.status_code != 200:
        print r.text
        return
    builders = json.loads(r.text)
    for builder_name in sorted(builders):
        # Build -1 is the currently-running build (if there is one), so we
        # start with -2, which should be the most or second-most
        # recently-completed build.
        for build_num in range(-2, -2 - NUM_BUILDS, -1):
            print ('Loading data for %s, build %d...'
                   % (builder_name, build_num))
            r = requests.get(
                'https://build.chromium.org/p/client.art'
                '/json/builders/%s/builds/%d' % (
                builder_name, build_num))
            if r.status_code != 200:
                print r.text
                return
            builder = json.loads(r.text)
            libcore_steps = [x for x in builder['steps']
                             if x['name'] == BUILDBOT_STEP]
            for ls in libcore_steps:
                stdio_logs = [x for x in ls['logs'] if x[0] == 'stdio']
                for sl in stdio_logs:
                    # The default link is HTML, so append /text to get the
                    # text version
                    r = requests.get(sl[1] + '/text')
                    if r.status_code != 200:
                        print r.text
                        return
                    stdio = r.text.splitlines()

                    # Walk from the back of the list to find the start of the
                    # warnings summary
                    i = -1
                    try:
                        while not stdio[i].startswith('Warnings summary:'):
                            i -= 1
                        i += 1   # Ignore the "Warnings summary:" line
                        while i < -1:
                            warnings['%s:%d' % (builder_name, build_num)].append(stdio[i])
                            i += 1
                    except IndexError:
                        # Some builds don't have any
                        print '  No warnings section found.'
    # sharedwarnings will build up the intersection of all the lists of
    # warnings.  We seed it with an arbitrary starting point (which is fine
    # since intersection is commutative).
    sharedwarnings = set(warnings.popitem()[1])
    for warning_list in warnings.itervalues():
        sharedwarnings = sharedwarnings & set(warning_list)
    print 'Warnings shared across all builders:'
    for warning in sorted(list(sharedwarnings)):
        print warning


if __name__ == '__main__':
    main()



#!/usr/bin/python3
#
# Copyright (C) 2015 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generate Java Main file from a classes.xml file.
"""

import os
import sys
from pathlib import Path

BUILD_TOP = os.getenv("ANDROID_BUILD_TOP")
if BUILD_TOP is None:
  print("ANDROID_BUILD_TOP not set. Please run build/envsetup.sh", file=sys.stderr)
  sys.exit(1)

# Allow us to import utils and mixins.
sys.path.append(str(Path(BUILD_TOP)/"art"/"test"/"utils"/"python"))

from testgen.utils import get_copyright
import testgen.mixins as mixins

from collections import namedtuple
import itertools
import functools
import xml.etree.ElementTree as ET

class MainClass(mixins.DumpMixin, mixins.Named, mixins.JavaFileMixin):
  """
  A mainclass and main method for this test.
  """

  MAIN_CLASS_TEMPLATE = """{copyright}
class Main {{
{test_groups}
{test_funcs}
{main_func}
}}
"""

  MAIN_FUNCTION_TEMPLATE = """
  public static void main(String[] args) {{
    {test_group_invoke}
  }}
"""

  TEST_GROUP_INVOKE_TEMPLATE = """
    {test_name}();
"""

  def __init__(self):
    """
    Initialize this MainClass
    """
    self.tests = set()
    self.global_funcs = set()

  def add_instance(self, it):
    """
    Add an instance test for the given class
    """
    self.tests.add(it)

  def add_func(self, f):
    """
    Add a function to the class
    """
    self.global_funcs.add(f)

  def get_name(self):
    """
    Get the name of this class
    """
    return "Main"

  def __str__(self):
    """
    Print this class
    """
    all_tests = sorted(self.tests)
    test_invoke = ""
    test_groups = ""
    for t in all_tests:
      test_groups += str(t)
    for t in sorted(all_tests):
      test_invoke += self.TEST_GROUP_INVOKE_TEMPLATE.format(test_name=t.get_name())
    main_func = self.MAIN_FUNCTION_TEMPLATE.format(test_group_invoke=test_invoke)

    funcs = ""
    for f in self.global_funcs:
      funcs += str(f)
    return self.MAIN_CLASS_TEMPLATE.format(copyright = get_copyright('java'),
                                           test_groups=test_groups,
                                           main_func=main_func, test_funcs=funcs)


class InstanceTest(mixins.Named, mixins.NameComparableMixin):
  """
  A method that runs tests for a particular concrete type, It calls the test
  cases for running it in all possible ways.
  """

  INSTANCE_TEST_TEMPLATE = """
  public static void {test_name}() {{
    System.out.println("Testing for type {ty}");
    String s = "{ty}";
    {ty} v = new {ty}();

    {invokes}

    System.out.println("End testing for type {ty}");
  }}
"""

  TEST_INVOKE_TEMPLATE = """
    {fname}(s, v);
"""

  def __init__(self, main, ty):
    """
    Initialize this test group for the given type
    """
    self.ty = ty
    self.main = main
    self.funcs = set()
    self.main.add_instance(self)

  def get_name(self):
    """
    Get the name of this test group
    """
    return "TEST_NAME_"+self.ty

  def add_func(self, f):
    """
    Add a test function to this test group
    """
    self.main.add_func(f)
    self.funcs.add(f)

  def __str__(self):
    """
    Returns the java code for this function
    """
    func_invokes = ""
    for f in sorted(self.funcs, key=lambda a: (a.func, a.farg)):
      func_invokes += self.TEST_INVOKE_TEMPLATE.format(fname=f.get_name(),
                                                       farg=f.farg)

    return self.INSTANCE_TEST_TEMPLATE.format(test_name=self.get_name(), ty=self.ty,
                                              invokes=func_invokes)

class Func(mixins.Named, mixins.NameComparableMixin):
  """
  A single test case that attempts to invoke a function on receiver of a given type.
  """

  TEST_FUNCTION_TEMPLATE = """
  public static void {fname}(String s, {farg} v) {{
    try {{
      System.out.printf("%s-{invoke_type:<9} {farg:>9}.{callfunc}()='%s'\\n", s, v.{callfunc}());
      return;
    }} catch (Error e) {{
      System.out.printf("%s-{invoke_type} on {farg}: {callfunc}() threw exception!\\n", s);
      if (e instanceof IncompatibleClassChangeError) {{
        System.out.printf("Exception is of type %s\\n", e.getClass().getName());
      }} else {{
        e.printStackTrace(System.out);
      }}
    }}
  }}
"""

  def __init__(self, func, farg, invoke):
    """
    Initialize this test function for the given invoke type and argument
    """
    self.func = func
    self.farg = farg
    self.invoke = invoke

  def get_name(self):
    """
    Get the name of this test
    """
    return "Test_Func_{}_{}_{}".format(self.func, self.farg, self.invoke)

  def __str__(self):
    """
    Get the java code for this test function
    """
    return self.TEST_FUNCTION_TEMPLATE.format(fname=self.get_name(),
                                              farg=self.farg,
                                              invoke_type=self.invoke,
                                              callfunc=self.func)

def flatten_classes(classes, c):
  """
  Iterate over all the classes 'c' can be used as
  """
  while c:
    yield c
    c = classes.get(c.super_class)

def flatten_class_methods(classes, c):
  """
  Iterate over all the methods 'c' can call
  """
  for c1 in flatten_classes(classes, c):
    yield from c1.methods

def flatten_interfaces(dat, c):
  """
  Iterate over all the interfaces 'c' transitively implements
  """
  def get_ifaces(cl):
    for i2 in cl.implements:
      yield dat.interfaces[i2]
      yield from get_ifaces(dat.interfaces[i2])

  for cl in flatten_classes(dat.classes, c):
    yield from get_ifaces(cl)

def flatten_interface_methods(dat, i):
  """
  Iterate over all the interface methods 'c' can call
  """
  yield from i.methods
  for i2 in flatten_interfaces(dat, i):
    yield from i2.methods

def make_main_class(dat):
  """
  Creates a Main.java file that runs all the tests
  """
  m = MainClass()
  for c in dat.classes.values():
    i = InstanceTest(m, c.name)
    for clazz in flatten_classes(dat.classes, c):
      for meth in flatten_class_methods(dat.classes, clazz):
        i.add_func(Func(meth, clazz.name, 'virtual'))
      for iface in flatten_interfaces(dat, clazz):
        for meth in flatten_interface_methods(dat, iface):
          i.add_func(Func(meth, clazz.name, 'virtual'))
          i.add_func(Func(meth, iface.name, 'interface'))
  return m

class TestData(namedtuple("TestData", ['classes', 'interfaces'])):
  """
  A class representing the classes.xml document.
  """
  pass

class Clazz(namedtuple("Clazz", ["name", "methods", "super_class", "implements"])):
  """
  A class representing a class element in the classes.xml document.
  """
  pass

class IFace(namedtuple("IFace", ["name", "methods", "super_class", "implements"])):
  """
  A class representing an interface element in the classes.xml document.
  """
  pass

def parse_xml(xml):
  """
  Parse the xml description of this test.
  """
  classes = dict()
  ifaces  = dict()
  root = ET.fromstring(xml)
  for iface in root.find("interfaces"):
    name = iface.attrib['name']
    implements = [a.text for a in iface.find("implements")]
    methods = [a.text for a in iface.find("methods")]
    ifaces[name] = IFace(name = name,
                         super_class = iface.attrib['super'],
                         methods = methods,
                         implements = implements)
  for clazz in root.find('classes'):
    name = clazz.attrib['name']
    implements = [a.text for a in clazz.find("implements")]
    methods = [a.text for a in clazz.find("methods")]
    classes[name] = Clazz(name = name,
                          super_class = clazz.attrib['super'],
                          methods = methods,
                          implements = implements)
  return TestData(classes, ifaces)

def main(argv):
  java_dir = Path(argv[1])
  if not java_dir.exists() or not java_dir.is_dir():
    print("{} is not a valid java dir".format(java_dir), file=sys.stderr)
    sys.exit(1)
  class_data = parse_xml((java_dir / "classes.xml").open().read())
  make_main_class(class_data).dump(java_dir)

if __name__ == '__main__':
  main(sys.argv)



# Copyright (C) 2014 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from common.logger import Logger
from common.mixins import EqualityMixin, PrintableMixin

import re

class CheckerFile(PrintableMixin):

  def __init__(self, fileName):
    self.fileName = fileName
    self.testCases = []

  def addTestCase(self, new_test_case):
    self.testCases.append(new_test_case)

  def testCasesForArch(self, targetArch):
    return [t for t in self.testCases if t.testArch == targetArch]

  def __eq__(self, other):
    return isinstance(other, self.__class__) \
       and self.testCases == other.testCases


class TestCase(PrintableMixin):

  def __init__(self, parent, name, startLineNo, testArch = None, forDebuggable = False):
    assert isinstance(parent, CheckerFile)

    self.parent = parent
    self.name = name
    self.assertions = []
    self.startLineNo = startLineNo
    self.testArch = testArch
    self.forDebuggable = forDebuggable

    if not self.name:
      Logger.fail("Test case does not have a name", self.fileName, self.startLineNo)

    self.parent.addTestCase(self)

  @property
  def fileName(self):
    return self.parent.fileName

  def addAssertion(self, new_assertion):
    if new_assertion.variant == TestAssertion.Variant.NextLine:
      if not self.assertions or \
         (self.assertions[-1].variant != TestAssertion.Variant.InOrder and \
          self.assertions[-1].variant != TestAssertion.Variant.NextLine):
        Logger.fail("A next-line assertion can only be placed after an "
                    "in-order assertion or another next-line assertion.",
                    new_assertion.fileName, new_assertion.lineNo)
    self.assertions.append(new_assertion)

  def __eq__(self, other):
    return isinstance(other, self.__class__) \
       and self.name == other.name \
       and self.assertions == other.assertions


class TestAssertion(PrintableMixin):

  class Variant(object):
    """Supported types of assertions."""
    InOrder, NextLine, DAG, Not, Eval = range(5)

  def __init__(self, parent, variant, originalText, lineNo):
    assert isinstance(parent, TestCase)

    self.parent = parent
    self.variant = variant
    self.expressions = []
    self.lineNo = lineNo
    self.originalText = originalText

    self.parent.addAssertion(self)

  @property
  def fileName(self):
    return self.parent.fileName

  def addExpression(self, new_expression):
    assert isinstance(new_expression, TestExpression)
    if self.variant == TestAssertion.Variant.Not:
      if new_expression.variant == TestExpression.Variant.VarDef:
        Logger.fail("CHECK-NOT lines cannot define variables", self.fileName, self.lineNo)
    self.expressions.append(new_expression)

  def toRegex(self):
    """ Returns a regex pattern for this entire assertion. Only used in tests. """
    regex = ""
    for expression in self.expressions:
      if expression.variant == TestExpression.Variant.Separator:
        regex = regex + ", "
      else:
        regex = regex + "(" + expression.text + ")"
    return regex

  def __eq__(self, other):
    return isinstance(other, self.__class__) \
       and self.variant == other.variant \
       and self.expressions == other.expressions


class TestExpression(EqualityMixin, PrintableMixin):

  class Variant(object):
    """Supported language constructs."""
    PlainText, Pattern, VarRef, VarDef, Separator = range(5)

  class Regex(object):
    rName = r"([a-zA-Z][a-zA-Z0-9]*)"
    rRegex = r"(.+?)"
    rPatternStartSym = r"(\{\{)"
    rPatternEndSym = r"(\}\})"
    rVariableStartSym = r"(<<)"
    rVariableEndSym = r"(>>)"
    rVariableSeparator = r"(:)"
    rVariableDefinitionBody = rName + rVariableSeparator + rRegex

    regexPattern = rPatternStartSym + rRegex + rPatternEndSym
    regexVariableReference = rVariableStartSym + rName + rVariableEndSym
    regexVariableDefinition = rVariableStartSym + rVariableDefinitionBody + rVariableEndSym

  def __init__(self, variant, name, text):
    self.variant = variant
    self.name = name
    self.text = text

  def __eq__(self, other):
    return isinstance(other, self.__class__) \
       and self.variant == other.variant \
       and self.name == other.name \
       and self.text == other.text

  @staticmethod
  def createSeparator():
    return TestExpression(TestExpression.Variant.Separator, None, None)

  @staticmethod
  def createPlainText(text):
    return TestExpression(TestExpression.Variant.PlainText, None, text)

  @staticmethod
  def createPatternFromPlainText(text):
    return TestExpression(TestExpression.Variant.Pattern, None, re.escape(text))

  @staticmethod
  def createPattern(pattern):
    return TestExpression(TestExpression.Variant.Pattern, None, pattern)

  @staticmethod
  def createVariableReference(name):
    assert re.match(TestExpression.Regex.rName, name)
    return TestExpression(TestExpression.Variant.VarRef, name, None)

  @staticmethod
  def createVariableDefinition(name, pattern):
    assert re.match(TestExpression.Regex.rName, name)
    return TestExpression(TestExpression.Variant.VarDef, name, pattern)



#! /usr/bin/python3
#
# Copyright (C) 2017 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generate Java test files for test 648-many-direct-methods.
"""

import os
import sys
from pathlib import Path

BUILD_TOP = os.getenv("ANDROID_BUILD_TOP")
if BUILD_TOP is None:
  print("ANDROID_BUILD_TOP not set. Please run build/envsetup.sh", file=sys.stderr)
  sys.exit(1)

# Allow us to import utils and mixins.
sys.path.append(str(Path(BUILD_TOP)/"art"/"test"/"utils"/"python"))

from testgen.utils import get_copyright, subtree_sizes, gensym, filter_blanks
import testgen.mixins as mixins

class MainClass(mixins.DumpMixin, mixins.Named, mixins.JavaFileMixin):
  """
  A Main.java file containing the Main class and the main function. It will run
  all the test functions we have.
  """

  MAIN_CLASS_TEMPLATE = """{copyright}
public class Main {{
{main_func}
{test_groups}

}}"""

  MAIN_FUNCTION_TEMPLATE = """
  public static void main(String[] args) {
    System.out.println("passed");
  }"""

  def __init__(self):
    """
    Initialize this MainClass. We start out with no tests.
    """
    self.tests = set()

  def add_test_method(self, num):
    """
    Add test method number 'num'
    """
    self.tests.add(TestMethod(num))

  def get_name(self):
    """
    Get the name of this class
    """
    return "Main"

  def __str__(self):
    """
    Print the MainClass Java code.
    """
    all_tests = sorted(self.tests)
    test_groups = ""
    for t in all_tests:
      test_groups += str(t)
    main_func = self.MAIN_FUNCTION_TEMPLATE

    return self.MAIN_CLASS_TEMPLATE.format(copyright = get_copyright("java"),
                                           main_func = main_func,
                                           test_groups = test_groups)

class TestMethod(mixins.Named, mixins.NameComparableMixin):
  """
  A function that represents a test method. Should only be
  constructed by MainClass.add_test_method.
  """

  TEST_FUNCTION_TEMPLATE = """
  public static void {fname}() {{}}"""

  def __init__(self, farg):
    """
    Initialize a test method for the given argument.
    """
    self.farg = farg

  def get_name(self):
    """
    Get the name of this test method.
    """
    return "method{:05d}".format(self.farg)

  def __str__(self):
    """
    Print the Java code of this test method.
    """
    return self.TEST_FUNCTION_TEMPLATE.format(fname=self.get_name())

# Number of generated test methods. This number has been chosen to
# make sure the number of direct methods in class Main is greater or
# equal to 2^16, and thus requires an *unsigned* 16-bit (short)
# integer to be represented (b/33650497).
NUM_TEST_METHODS = 32768

def create_test_file():
  """
  Creates the object representing the test file. It just needs to be dumped.
  """
  mc = MainClass()
  for i in range(1, NUM_TEST_METHODS + 1):
    mc.add_test_method(i)
  return mc

def main(argv):
  java_dir = Path(argv[1])
  if not java_dir.exists() or not java_dir.is_dir():
    print("{} is not a valid Java dir".format(java_dir), file=sys.stderr)
    sys.exit(1)
  mainclass = create_test_file()
  mainclass.dump(java_dir)

if __name__ == '__main__':
  main(sys.argv)



#!/usr/bin/env python3
#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import argparse
import filecmp
import os
import shlex
import shutil
import subprocess
import sys

from glob import glob
from subprocess import DEVNULL
from tempfile import mkdtemp

sys.path.append(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__))))

from common.common import RetCode
from common.common import CommandListToCommandString
from common.common import FatalError
from common.common import GetEnvVariableOrError
from common.common import RunCommand
from common.common import RunCommandForOutput
from common.common import DeviceTestEnv

# Return codes supported by bisection bug search.
BISECTABLE_RET_CODES = (RetCode.SUCCESS, RetCode.ERROR, RetCode.TIMEOUT)


def GetExecutionModeRunner(dexer, debug_info, device, mode):
  """Returns a runner for the given execution mode.

  Args:
    dexer: string, defines dexer
    debug_info: boolean, if True include debugging info
    device: string, target device serial number (or None)
    mode: string, execution mode
  Returns:
    TestRunner with given execution mode
  Raises:
    FatalError: error for unknown execution mode
  """
  if mode == 'ri':
    return TestRunnerRIOnHost(debug_info)
  if mode == 'hint':
    return TestRunnerArtIntOnHost(dexer, debug_info)
  if mode == 'hopt':
    return TestRunnerArtOptOnHost(dexer, debug_info)
  if mode == 'tint':
    return TestRunnerArtIntOnTarget(dexer, debug_info, device)
  if mode == 'topt':
    return TestRunnerArtOptOnTarget(dexer, debug_info, device)
  raise FatalError('Unknown execution mode')


#
# Execution mode classes.
#


class TestRunner(object):
  """Abstraction for running a test in a particular execution mode."""
  __meta_class__ = abc.ABCMeta

  @abc.abstractproperty
  def description(self):
    """Returns a description string of the execution mode."""

  @abc.abstractproperty
  def id(self):
    """Returns a short string that uniquely identifies the execution mode."""

  @property
  def output_file(self):
    return self.id + '_out.txt'

  @abc.abstractmethod
  def GetBisectionSearchArgs(self):
    """Get arguments to pass to bisection search tool.

    Returns:
      list of strings - arguments for bisection search tool, or None if
      runner is not bisectable
    """

  @abc.abstractmethod
  def CompileAndRunTest(self):
    """Compile and run the generated test.

    Ensures that the current Test.java in the temporary directory is compiled
    and executed under the current execution mode. On success, transfers the
    generated output to the file self.output_file in the temporary directory.

    Most nonzero return codes are assumed non-divergent, since systems may
    exit in different ways. This is enforced by normalizing return codes.

    Returns:
      normalized return code
    """


class TestRunnerWithHostCompilation(TestRunner):
  """Abstract test runner that supports compilation on host."""

  def  __init__(self, dexer, debug_info):
    """Constructor for the runner with host compilation.

    Args:
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
    """
    self._dexer = dexer
    self._debug_info = debug_info

  def CompileOnHost(self):
    if self._dexer == 'dx' or self._dexer == 'd8':
      dbg = '-g' if self._debug_info else '-g:none'
      if RunCommand(['javac', '--release=8', dbg, 'Test.java'],
                    out=None, err=None, timeout=30) == RetCode.SUCCESS:
        dx = 'dx' if self._dexer == 'dx' else 'd8-compat-dx'
        retc = RunCommand([dx, '--dex', '--output=classes.dex'] + glob('*.class'),
                          out=None, err='dxerr.txt', timeout=30)
      else:
        retc = RetCode.NOTCOMPILED
    else:
      raise FatalError('Unknown dexer: ' + self._dexer)
    return retc


class TestRunnerRIOnHost(TestRunner):
  """Concrete test runner of the reference implementation on host."""

  def  __init__(self, debug_info):
    """Constructor for the runner with host compilation.

    Args:
      debug_info: boolean, if True include debugging info
    """
    self._debug_info = debug_info

  @property
  def description(self):
    return 'RI on host'

  @property
  def id(self):
    return 'RI'

  def CompileAndRunTest(self):
    dbg = '-g' if self._debug_info else '-g:none'
    if RunCommand(['javac', '--release=8', dbg, 'Test.java'],
                  out=None, err=None, timeout=30) == RetCode.SUCCESS:
      retc = RunCommand(['java', 'Test'], self.output_file, err=None)
    else:
      retc = RetCode.NOTCOMPILED
    return retc

  def GetBisectionSearchArgs(self):
    return None


class TestRunnerArtOnHost(TestRunnerWithHostCompilation):
  """Abstract test runner of Art on host."""

  def  __init__(self, dexer, debug_info, extra_args=None):
    """Constructor for the Art on host tester.

    Args:
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
      extra_args: list of strings, extra arguments for dalvikvm
    """
    super().__init__(dexer, debug_info)
    self._art_cmd = ['/bin/bash', 'art', '-cp', 'classes.dex']
    if extra_args is not None:
      self._art_cmd += extra_args
    self._art_cmd.append('Test')

  def CompileAndRunTest(self):
    if self.CompileOnHost() == RetCode.SUCCESS:
      retc = RunCommand(self._art_cmd, self.output_file, 'arterr.txt')
    else:
      retc = RetCode.NOTCOMPILED
    return retc


class TestRunnerArtIntOnHost(TestRunnerArtOnHost):
  """Concrete test runner of interpreter mode Art on host."""

  def  __init__(self, dexer, debug_info):
    """Constructor for the Art on host tester (interpreter).

    Args:
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
   """
    super().__init__(dexer, debug_info, ['-Xint'])

  @property
  def description(self):
    return 'Art interpreter on host'

  @property
  def id(self):
    return 'HInt'

  def GetBisectionSearchArgs(self):
    return None


class TestRunnerArtOptOnHost(TestRunnerArtOnHost):
  """Concrete test runner of optimizing compiler mode Art on host."""

  def  __init__(self, dexer, debug_info):
    """Constructor for the Art on host tester (optimizing).

    Args:
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
   """
    super().__init__(dexer, debug_info, None)

  @property
  def description(self):
    return 'Art optimizing on host'

  @property
  def id(self):
    return 'HOpt'

  def GetBisectionSearchArgs(self):
    cmd_str = CommandListToCommandString(
        self._art_cmd[0:2] + ['{ARGS}'] + self._art_cmd[2:])
    return ['--raw-cmd={0}'.format(cmd_str), '--timeout', str(30)]


class TestRunnerArtOnTarget(TestRunnerWithHostCompilation):
  """Abstract test runner of Art on target."""

  def  __init__(self, dexer, debug_info, device, extra_args=None):
    """Constructor for the Art on target tester.

    Args:
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
      device: string, target device serial number (or None)
      extra_args: list of strings, extra arguments for dalvikvm
    """
    super().__init__(dexer, debug_info)
    self._test_env = DeviceTestEnv('jfuzz_', specific_device=device)
    self._dalvik_cmd = ['dalvikvm']
    if extra_args is not None:
      self._dalvik_cmd += extra_args
    self._device = device
    self._device_classpath = None

  def CompileAndRunTest(self):
    if self.CompileOnHost() == RetCode.SUCCESS:
      self._device_classpath = self._test_env.PushClasspath('classes.dex')
      cmd = self._dalvik_cmd + ['-cp', self._device_classpath, 'Test']
      (output, retc) = self._test_env.RunCommand(
          cmd, {'ANDROID_LOG_TAGS': '*:s'})
      with open(self.output_file, 'w') as run_out:
        run_out.write(output)
    else:
      retc = RetCode.NOTCOMPILED
    return retc

  def GetBisectionSearchArgs(self):
    cmd_str = CommandListToCommandString(
        self._dalvik_cmd + ['-cp',self._device_classpath, 'Test'])
    cmd = ['--raw-cmd={0}'.format(cmd_str), '--timeout', str(30)]
    if self._device:
      cmd += ['--device-serial', self._device]
    else:
      cmd.append('--device')
    return cmd


class TestRunnerArtIntOnTarget(TestRunnerArtOnTarget):
  """Concrete test runner of interpreter mode Art on target."""

  def  __init__(self, dexer, debug_info, device):
    """Constructor for the Art on target tester (interpreter).

    Args:
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
      device: string, target device serial number (or None)
    """
    super().__init__(dexer, debug_info, device, ['-Xint'])

  @property
  def description(self):
    return 'Art interpreter on target'

  @property
  def id(self):
    return 'TInt'

  def GetBisectionSearchArgs(self):
    return None


class TestRunnerArtOptOnTarget(TestRunnerArtOnTarget):
  """Concrete test runner of optimizing compiler mode Art on target."""

  def  __init__(self, dexer, debug_info, device):
    """Constructor for the Art on target tester (optimizing).

    Args:
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
      device: string, target device serial number (or None)
    """
    super().__init__(dexer, debug_info, device, None)

  @property
  def description(self):
    return 'Art optimizing on target'

  @property
  def id(self):
    return 'TOpt'

  def GetBisectionSearchArgs(self):
    cmd_str = CommandListToCommandString(
        self._dalvik_cmd + ['-cp', self._device_classpath, 'Test'])
    cmd = ['--raw-cmd={0}'.format(cmd_str), '--timeout', str(30)]
    if self._device:
      cmd += ['--device-serial', self._device]
    else:
      cmd.append('--device')
    return cmd


#
# Tester class.
#


class JFuzzTester(object):
  """Tester that runs JFuzz many times and report divergences."""

  def  __init__(self, num_tests, device, mode1, mode2, jfuzz_args,
                report_script, true_divergence_only, dexer, debug_info):
    """Constructor for the tester.

    Args:
      num_tests: int, number of tests to run
      device: string, target device serial number (or None)
      mode1: string, execution mode for first runner
      mode2: string, execution mode for second runner
      jfuzz_args: list of strings, additional arguments for jfuzz
      report_script: string, path to script called for each divergence
      true_divergence_only: boolean, if True don't bisect timeout divergences
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
    """
    self._num_tests = num_tests
    self._device = device
    self._runner1 = GetExecutionModeRunner(dexer, debug_info, device, mode1)
    self._runner2 = GetExecutionModeRunner(dexer, debug_info, device, mode2)
    self._jfuzz_args = jfuzz_args
    self._report_script = report_script
    self._true_divergence_only = true_divergence_only
    self._dexer = dexer
    self._debug_info = debug_info
    self._save_dir = None
    self._results_dir = None
    self._jfuzz_dir = None
    # Statistics.
    self._test = 0
    self._num_success = 0
    self._num_not_compiled = 0
    self._num_not_run = 0
    self._num_timed_out = 0
    self._num_divergences = 0

  def __enter__(self):
    """On entry, enters new temp directory after saving current directory.

    Raises:
      FatalError: error when temp directory cannot be constructed
    """
    self._save_dir = os.getcwd()
    self._results_dir = mkdtemp(dir='/tmp/')
    self._jfuzz_dir = mkdtemp(dir=self._results_dir)
    if self._results_dir is None or self._jfuzz_dir is None:
      raise FatalError('Cannot obtain temp directory')
    os.chdir(self._jfuzz_dir)
    return self

  def __exit__(self, etype, evalue, etraceback):
    """On exit, re-enters previously saved current directory and cleans up."""
    os.chdir(self._save_dir)
    shutil.rmtree(self._jfuzz_dir)
    if self._num_divergences == 0:
      shutil.rmtree(self._results_dir)

  def Run(self):
    """Runs JFuzz many times and report divergences."""
    print()
    print('**\n**** JFuzz Testing\n**')
    print()
    print('#Tests    :', self._num_tests)
    print('Device    :', self._device)
    print('Directory :', self._results_dir)
    print('Exec-mode1:', self._runner1.description)
    print('Exec-mode2:', self._runner2.description)
    print('Dexer     :', self._dexer)
    print('Debug-info:', self._debug_info)
    print()
    self.ShowStats()
    for self._test in range(1, self._num_tests + 1):
      self.RunJFuzzTest()
      self.ShowStats()
    if self._num_divergences == 0:
      print('\n\nsuccess (no divergences)\n')
    else:
      print('\n\nfailure (divergences)\n')

  def ShowStats(self):
    """Shows current statistics (on same line) while tester is running."""
    print('\rTests:', self._test,
          'Success:', self._num_success,
          'Not-compiled:', self._num_not_compiled,
          'Not-run:', self._num_not_run,
          'Timed-out:', self._num_timed_out,
          'Divergences:', self._num_divergences,
          end='')
    sys.stdout.flush()

  def RunJFuzzTest(self):
    """Runs a single JFuzz test, comparing two execution modes."""
    self.ConstructTest()
    retc1 = self._runner1.CompileAndRunTest()
    retc2 = self._runner2.CompileAndRunTest()
    self.CheckForDivergence(retc1, retc2)
    self.CleanupTest()

  def ConstructTest(self):
    """Use JFuzz to generate next Test.java test.

    Raises:
      FatalError: error when jfuzz fails
    """
    if (RunCommand(['jfuzz'] + self._jfuzz_args, out='Test.java', err=None)
          != RetCode.SUCCESS):
      raise FatalError('Unexpected error while running JFuzz')

  def CheckForDivergence(self, retc1, retc2):
    """Checks for divergences and updates statistics.

    Args:
      retc1: int, normalized return code of first runner
      retc2: int, normalized return code of second runner
    """
    if retc1 == retc2:
      # No divergence in return code.
      if retc1 == RetCode.SUCCESS:
        # Both compilations and runs were successful, inspect generated output.
        runner1_out = self._runner1.output_file
        runner2_out = self._runner2.output_file
        if not filecmp.cmp(runner1_out, runner2_out, shallow=False):
          # Divergence in output.
          self.ReportDivergence(retc1, retc2, is_output_divergence=True)
        else:
          # No divergence in output.
          self._num_success += 1
      elif retc1 == RetCode.TIMEOUT:
        self._num_timed_out += 1
      elif retc1 == RetCode.NOTCOMPILED:
        self._num_not_compiled += 1
      else:
        self._num_not_run += 1
    else:
      # Divergence in return code.
      if self._true_divergence_only:
        # When only true divergences are requested, any divergence in return
        # code where one is a time out is treated as a regular time out.
        if RetCode.TIMEOUT in (retc1, retc2):
          self._num_timed_out += 1
          return
        # When only true divergences are requested, a runtime crash in just
        # the RI is treated as if not run at all.
        if retc1 == RetCode.ERROR and retc2 == RetCode.SUCCESS:
          if self._runner1.GetBisectionSearchArgs() is None:
            self._num_not_run += 1
            return
      self.ReportDivergence(retc1, retc2, is_output_divergence=False)

  def GetCurrentDivergenceDir(self):
    return self._results_dir + '/divergence' + str(self._num_divergences)

  def ReportDivergence(self, retc1, retc2, is_output_divergence):
    """Reports and saves a divergence."""
    self._num_divergences += 1
    print('\n' + str(self._num_divergences), end='')
    if is_output_divergence:
      print(' divergence in output')
    else:
      print(' divergence in return code: ' + retc1.name + ' vs. ' +
            retc2.name)
    # Save.
    ddir = self.GetCurrentDivergenceDir()
    os.mkdir(ddir)
    for f in glob('*.txt') + ['Test.java']:
      shutil.copy(f, ddir)
    # Maybe run bisection bug search.
    if retc1 in BISECTABLE_RET_CODES and retc2 in BISECTABLE_RET_CODES:
      self.MaybeBisectDivergence(retc1, retc2, is_output_divergence)
    # Call reporting script.
    if self._report_script:
      self.RunReportScript(retc1, retc2, is_output_divergence)

  def RunReportScript(self, retc1, retc2, is_output_divergence):
    """Runs report script."""
    try:
      title = "Divergence between {0} and {1} (found with fuzz testing)".format(
          self._runner1.description, self._runner2.description)
      # Prepare divergence comment.
      jfuzz_cmd_and_version = subprocess.check_output(
          ['grep', '-o', 'jfuzz.*', 'Test.java'], universal_newlines=True)
      (jfuzz_cmd_str, jfuzz_ver) = jfuzz_cmd_and_version.split('(')
      # Strip right parenthesis and new line.
      jfuzz_ver = jfuzz_ver[:-2]
      jfuzz_args = ['\'-{0}\''.format(arg)
                    for arg in jfuzz_cmd_str.strip().split(' -')][1:]
      wrapped_args = ['--jfuzz_arg={0}'.format(opt) for opt in jfuzz_args]
      repro_cmd_str = (os.path.basename(__file__) +
                       ' --num_tests=1 --dexer=' + self._dexer +
                       (' --debug_info ' if self._debug_info else ' ') +
                       ' '.join(wrapped_args))
      comment = 'jfuzz {0}\nReproduce test:\n{1}\nReproduce divergence:\n{2}\n'.format(
          jfuzz_ver, jfuzz_cmd_str, repro_cmd_str)
      if is_output_divergence:
        (output, _, _) = RunCommandForOutput(
            ['diff', self._runner1.output_file, self._runner2.output_file],
            None, subprocess.PIPE, subprocess.STDOUT)
        comment += 'Diff:\n' + output
      else:
        comment += '{0} vs {1}\n'.format(retc1, retc2)
      # Prepare report script command.
      script_cmd = [self._report_script, title, comment]
      ddir = self.GetCurrentDivergenceDir()
      bisection_out_files = glob(ddir + '/*_bisection_out.txt')
      if bisection_out_files:
        script_cmd += ['--bisection_out', bisection_out_files[0]]
      subprocess.check_call(script_cmd, stdout=DEVNULL, stderr=DEVNULL)
    except subprocess.CalledProcessError as err:
      print('Failed to run report script.\n', err)

  def RunBisectionSearch(self, args, expected_retcode, expected_output,
                         runner_id):
    ddir = self.GetCurrentDivergenceDir()
    outfile_path = ddir + '/' + runner_id + '_bisection_out.txt'
    logfile_path = ddir + '/' + runner_id + '_bisection_log.txt'
    errfile_path = ddir + '/' + runner_id + '_bisection_err.txt'
    args = list(args) + ['--logfile', logfile_path, '--cleanup']
    args += ['--expected-retcode', expected_retcode.name]
    if expected_output:
      args += ['--expected-output', expected_output]
    bisection_search_path = os.path.join(
        GetEnvVariableOrError('ANDROID_BUILD_TOP'),
        'art/tools/bisection_search/bisection_search.py')
    if RunCommand([bisection_search_path] + args, out=outfile_path,
                  err=errfile_path, timeout=300) == RetCode.TIMEOUT:
      print('Bisection search TIMEOUT')

  def MaybeBisectDivergence(self, retc1, retc2, is_output_divergence):
    bisection_args1 = self._runner1.GetBisectionSearchArgs()
    bisection_args2 = self._runner2.GetBisectionSearchArgs()
    if is_output_divergence:
      maybe_output1 = self._runner1.output_file
      maybe_output2 = self._runner2.output_file
    else:
      maybe_output1 = maybe_output2 = None
    if bisection_args1 is not None:
      self.RunBisectionSearch(bisection_args1, retc2, maybe_output2,
                              self._runner1.id)
    if bisection_args2 is not None:
      self.RunBisectionSearch(bisection_args2, retc1, maybe_output1,
                              self._runner2.id)

  def CleanupTest(self):
    """Cleans up after a single test run."""
    for file_name in os.listdir(self._jfuzz_dir):
      file_path = os.path.join(self._jfuzz_dir, file_name)
      if os.path.isfile(file_path):
        os.unlink(file_path)
      elif os.path.isdir(file_path):
        shutil.rmtree(file_path)


def main():
  # Handle arguments.
  parser = argparse.ArgumentParser()
  parser.add_argument('--num_tests', default=10000, type=int,
                      help='number of tests to run')
  parser.add_argument('--device', help='target device serial number')
  parser.add_argument('--mode1', default='ri',
                      help='execution mode 1 (default: ri)')
  parser.add_argument('--mode2', default='hopt',
                      help='execution mode 2 (default: hopt)')
  parser.add_argument('--report_script',
                      help='script called for each divergence')
  parser.add_argument('--jfuzz_arg', default=[], dest='jfuzz_args',
                      action='append',
                      help='argument for jfuzz')
  parser.add_argument('--true_divergence', default=False, action='store_true',
                      help='do not bisect timeout divergences')
  parser.add_argument('--dexer', default='dx', type=str,
                      help='defines dexer as dx or d8 (default: dx)')
  parser.add_argument('--debug_info', default=False, action='store_true',
                      help='include debugging info')
  args = parser.parse_args()
  if args.mode1 == args.mode2:
    raise FatalError('Identical execution modes given')
  # Run the JFuzz tester.
  with JFuzzTester(args.num_tests,
                   args.device,
                   args.mode1, args.mode2,
                   args.jfuzz_args,
                   args.report_script,
                   args.true_divergence,
                   args.dexer,
                   args.debug_info) as fuzzer:
    fuzzer.Run()

if __name__ == '__main__':
  main()



#!/usr/bin/env python3
#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import re
import shutil
import subprocess
import sys

from glob import glob

from tempfile import mkdtemp
from tempfile import TemporaryFile

# run_jfuzz_test.py success/failure strings.
SUCCESS_STRING = 'success (no divergences)'
FAILURE_STRING = 'FAILURE (divergences)'

# Constant returned by string find() method when search fails.
NOT_FOUND = -1

def main(argv):
  # Set up.
  cwd = os.path.dirname(os.path.realpath(__file__))
  cmd = [cwd + '/run_jfuzz_test.py']
  parser = argparse.ArgumentParser()
  parser.add_argument('--num_proc', default=8,
                      type=int, help='number of processes to run')
  # Unknown arguments are passed to run_jfuzz_test.py.
  (args, unknown_args) = parser.parse_known_args()
  # Run processes.
  cmd = cmd + unknown_args
  print()
  print('**\n**** Nightly JFuzz Testing\n**')
  print()
  print('**** Running ****\n\n', cmd, '\n')
  output_files = [TemporaryFile('wb+') for _ in range(args.num_proc)]
  processes = []
  for i, output_file in enumerate(output_files):
    print('Tester', i)
    processes.append(subprocess.Popen(cmd, stdout=output_file,
                                      stderr=subprocess.STDOUT))
  try:
    # Wait for processes to terminate.
    for proc in processes:
      proc.wait()
  except KeyboardInterrupt:
    for proc in processes:
      proc.kill()
  # Output results.
  print('\n**** Results ****\n')
  output_dirs = []
  for i, output_file in enumerate(output_files):
    output_file.seek(0)
    output_str = output_file.read().decode('ascii')
    output_file.close()
    # Extract output directory. Example match: 'Directory : /tmp/tmp8ltpfjng'.
    directory_match = re.search(r'Directory[^:]*: ([^\n]+)\n', output_str)
    if directory_match:
      output_dirs.append(directory_match.group(1))
    if output_str.find(SUCCESS_STRING) == NOT_FOUND:
      print('Tester', i, FAILURE_STRING)
    else:
      print('Tester', i, SUCCESS_STRING)
  # Gather divergences.
  global_out_dir = mkdtemp('jfuzz_nightly')
  divergence_nr = 0
  for out_dir in output_dirs:
    for divergence_dir in glob(out_dir + '/divergence*/'):
      divergence_nr += 1
      shutil.copytree(divergence_dir,
                      global_out_dir + '/divergence' + str(divergence_nr))
  if divergence_nr > 0:
    print('\n!!!! Divergences !!!!', divergence_nr)
  else:
    print ('\nSuccess')
  print('\nGlobal output directory:', global_out_dir)
  print()

if __name__ == '__main__':
  main(sys.argv)



#!/usr/bin/env python2
#
# Copyright (C) 2014 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from common.archs               import archs_list
from common.testing             import ToUnicode
from file_format.checker.parser import ParseCheckerStream
from file_format.checker.struct import CheckerFile, TestCase, TestAssertion, TestExpression

import io
import unittest

CheckerException = SystemExit

class CheckerParser_PrefixTest(unittest.TestCase):

  def tryParse(self, string):
    checkerText = u"/// CHECK-START: pass\n" + ToUnicode(string)
    return ParseCheckerStream("<test-file>", "CHECK", io.StringIO(checkerText))

  def assertParses(self, string):
    checkFile = self.tryParse(string)
    self.assertEqual(len(checkFile.testCases), 1)
    self.assertNotEqual(len(checkFile.testCases[0].assertions), 0)

  def assertIgnored(self, string):
    checkFile = self.tryParse(string)
    self.assertEqual(len(checkFile.testCases), 1)
    self.assertEqual(len(checkFile.testCases[0].assertions), 0)

  def assertInvalid(self, string):
    with self.assertRaises(CheckerException):
      self.tryParse(string)

  def test_ValidFormat(self):
    self.assertParses("///CHECK:foo")
    self.assertParses("##CHECK:bar")

  def test_InvalidFormat(self):
    self.assertIgnored("CHECK")
    self.assertIgnored(":CHECK")
    self.assertIgnored("CHECK:")
    self.assertIgnored("//CHECK")
    self.assertIgnored("#CHECK")
    self.assertInvalid("///CHECK")
    self.assertInvalid("##CHECK")

  def test_InvalidPrefix(self):
    self.assertInvalid("///ACHECK:foo")
    self.assertInvalid("##ACHECK:foo")

  def test_NotFirstOnTheLine(self):
    self.assertIgnored("A/// CHECK: foo")
    self.assertIgnored("A # CHECK: foo")
    self.assertInvalid("/// /// CHECK: foo")
    self.assertInvalid("## ## CHECK: foo")

  def test_WhitespaceAgnostic(self):
    self.assertParses("  ///CHECK: foo")
    self.assertParses("///  CHECK: foo")
    self.assertParses("    ///CHECK: foo")
    self.assertParses("///    CHECK: foo")

class CheckerParser_TestExpressionTest(unittest.TestCase):

  def parseAssertion(self, string, variant=""):
    checkerText = (u"/// CHECK-START: pass\n" +
                   u"/// CHECK" + ToUnicode(variant) + u": " + ToUnicode(string))
    checkerFile = ParseCheckerStream("<test-file>", "CHECK", io.StringIO(checkerText))
    self.assertEqual(len(checkerFile.testCases), 1)
    testCase = checkerFile.testCases[0]
    self.assertEqual(len(testCase.assertions), 1)
    return testCase.assertions[0]

  def parseExpression(self, string):
    line = self.parseAssertion(string)
    self.assertEqual(1, len(line.expressions))
    return line.expressions[0]

  def assertEqualsRegex(self, string, expected):
    self.assertEqual(expected, self.parseAssertion(string).toRegex())

  def assertEqualsText(self, string, text):
    self.assertEqual(self.parseExpression(string), TestExpression.createPatternFromPlainText(text))

  def assertEqualsPattern(self, string, pattern):
    self.assertEqual(self.parseExpression(string), TestExpression.createPattern(pattern))

  def assertEqualsVarRef(self, string, name):
    self.assertEqual(self.parseExpression(string), TestExpression.createVariableReference(name))

  def assertEqualsVarDef(self, string, name, pattern):
    self.assertEqual(self.parseExpression(string),
                     TestExpression.createVariableDefinition(name, pattern))

  def assertVariantNotEqual(self, string, variant):
    self.assertNotEqual(variant, self.parseExpression(string).variant)

  # Test that individual parts of the line are recognized

  def test_TextOnly(self):
    self.assertEqualsText("foo", "foo")
    self.assertEqualsText("  foo  ", "foo")
    self.assertEqualsRegex("f$o^o", "(f\$o\^o)")

  def test_PatternOnly(self):
    self.assertEqualsPattern("{{a?b.c}}", "a?b.c")

  def test_VarRefOnly(self):
    self.assertEqualsVarRef("<<ABC>>", "ABC")

  def test_VarDefOnly(self):
    self.assertEqualsVarDef("<<ABC:a?b.c>>", "ABC", "a?b.c")

  def test_TextWithWhitespace(self):
    self.assertEqualsRegex("foo bar", "(foo), (bar)")
    self.assertEqualsRegex("foo   bar", "(foo), (bar)")

  def test_TextWithRegex(self):
    self.assertEqualsRegex("foo{{abc}}bar", "(foo)(abc)(bar)")

  def test_TextWithVar(self):
    self.assertEqualsRegex("foo<<ABC:abc>>bar", "(foo)(abc)(bar)")

  def test_PlainWithRegexAndWhitespaces(self):
    self.assertEqualsRegex("foo {{abc}}bar", "(foo), (abc)(bar)")
    self.assertEqualsRegex("foo{{abc}} bar", "(foo)(abc), (bar)")
    self.assertEqualsRegex("foo {{abc}} bar", "(foo), (abc), (bar)")

  def test_PlainWithVarAndWhitespaces(self):
    self.assertEqualsRegex("foo <<ABC:abc>>bar", "(foo), (abc)(bar)")
    self.assertEqualsRegex("foo<<ABC:abc>> bar", "(foo)(abc), (bar)")
    self.assertEqualsRegex("foo <<ABC:abc>> bar", "(foo), (abc), (bar)")

  def test_AllKinds(self):
    self.assertEqualsRegex("foo <<ABC:abc>>{{def}}bar", "(foo), (abc)(def)(bar)")
    self.assertEqualsRegex("foo<<ABC:abc>> {{def}}bar", "(foo)(abc), (def)(bar)")
    self.assertEqualsRegex("foo <<ABC:abc>> {{def}} bar", "(foo), (abc), (def), (bar)")

  # # Test that variables and patterns are parsed correctly

  def test_ValidPattern(self):
    self.assertEqualsPattern("{{abc}}", "abc")
    self.assertEqualsPattern("{{a[b]c}}", "a[b]c")
    self.assertEqualsPattern("{{(a{bc})}}", "(a{bc})")

  def test_ValidRef(self):
    self.assertEqualsVarRef("<<ABC>>", "ABC")
    self.assertEqualsVarRef("<<A1BC2>>", "A1BC2")

  def test_ValidDef(self):
    self.assertEqualsVarDef("<<ABC:abc>>", "ABC", "abc")
    self.assertEqualsVarDef("<<ABC:ab:c>>", "ABC", "ab:c")
    self.assertEqualsVarDef("<<ABC:a[b]c>>", "ABC", "a[b]c")
    self.assertEqualsVarDef("<<ABC:(a[bc])>>", "ABC", "(a[bc])")

  def test_Empty(self):
    self.assertEqualsText("{{}}", "{{}}")
    self.assertVariantNotEqual("<<>>", TestExpression.Variant.VarRef)
    self.assertVariantNotEqual("<<:>>", TestExpression.Variant.VarDef)

  def test_InvalidVarName(self):
    self.assertVariantNotEqual("<<0ABC>>", TestExpression.Variant.VarRef)
    self.assertVariantNotEqual("<<AB=C>>", TestExpression.Variant.VarRef)
    self.assertVariantNotEqual("<<ABC=>>", TestExpression.Variant.VarRef)
    self.assertVariantNotEqual("<<0ABC:abc>>", TestExpression.Variant.VarDef)
    self.assertVariantNotEqual("<<AB=C:abc>>", TestExpression.Variant.VarDef)
    self.assertVariantNotEqual("<<ABC=:abc>>", TestExpression.Variant.VarDef)

  def test_BodyMatchNotGreedy(self):
    self.assertEqualsRegex("{{abc}}{{def}}", "(abc)(def)")
    self.assertEqualsRegex("<<ABC:abc>><<DEF:def>>", "(abc)(def)")

  def test_NoVarDefsInNotChecks(self):
    with self.assertRaises(CheckerException):
      self.parseAssertion("<<ABC:abc>>", "-NOT")


class CheckerParser_FileLayoutTest(unittest.TestCase):

  # Creates an instance of CheckerFile from provided info.
  # Data format: [ ( <case-name>, [ ( <text>, <assert-variant> ), ... ] ), ... ]
  def createFile(self, caseList):
    testFile = CheckerFile("<test_file>")
    for caseEntry in caseList:
      caseName = caseEntry[0]
      testCase = TestCase(testFile, caseName, 0)
      assertionList = caseEntry[1]
      for assertionEntry in assertionList:
        content = assertionEntry[0]
        variant = assertionEntry[1]
        assertion = TestAssertion(testCase, variant, content, 0)
        assertion.addExpression(TestExpression.createPatternFromPlainText(content))
    return testFile

  def assertParsesTo(self, checkerText, expectedData):
    expectedFile = self.createFile(expectedData)
    actualFile = self.parse(checkerText)
    return self.assertEqual(expectedFile, actualFile)

  def parse(self, checkerText):
    return ParseCheckerStream("<test_file>", "CHECK", io.StringIO(ToUnicode(checkerText)))

  def test_EmptyFile(self):
    self.assertParsesTo("", [])

  def test_SingleGroup(self):
    self.assertParsesTo(
      """
        /// CHECK-START: Example Group
        /// CHECK:  foo
        /// CHECK:    bar
      """,
      [ ( "Example Group", [ ("foo", TestAssertion.Variant.InOrder),
                             ("bar", TestAssertion.Variant.InOrder) ] ) ])

  def test_MultipleGroups(self):
    self.assertParsesTo(
      """
        /// CHECK-START: Example Group1
        /// CHECK: foo
        /// CHECK: bar
        /// CHECK-START: Example Group2
        /// CHECK: abc
        /// CHECK: def
      """,
      [ ( "Example Group1", [ ("foo", TestAssertion.Variant.InOrder),
                              ("bar", TestAssertion.Variant.InOrder) ] ),
        ( "Example Group2", [ ("abc", TestAssertion.Variant.InOrder),
                              ("def", TestAssertion.Variant.InOrder) ] ) ])

  def test_AssertionVariants(self):
    self.assertParsesTo(
      """
        /// CHECK-START: Example Group
        /// CHECK:      foo1
        /// CHECK:      foo2
        /// CHECK-NEXT: foo3
        /// CHECK-NEXT: foo4
        /// CHECK-NOT:  bar
        /// CHECK-DAG:  abc
        /// CHECK-DAG:  def
      """,
      [ ( "Example Group", [ ("foo1", TestAssertion.Variant.InOrder),
                             ("foo2", TestAssertion.Variant.InOrder),
                             ("foo3", TestAssertion.Variant.NextLine),
                             ("foo4", TestAssertion.Variant.NextLine),
                             ("bar", TestAssertion.Variant.Not),
                             ("abc", TestAssertion.Variant.DAG),
                             ("def", TestAssertion.Variant.DAG) ] ) ])

  def test_MisplacedNext(self):
    with self.assertRaises(CheckerException):
      self.parse(
        """
          /// CHECK-START: Example Group
          /// CHECK-DAG:  foo
          /// CHECK-NEXT: bar
        """)
    with self.assertRaises(CheckerException):
      self.parse(
        """
          /// CHECK-START: Example Group
          /// CHECK-NOT:  foo
          /// CHECK-NEXT: bar
        """)
    with self.assertRaises(CheckerException):
      self.parse(
        """
          /// CHECK-START: Example Group
          /// CHECK-EVAL: foo
          /// CHECK-NEXT: bar
        """)
    with self.assertRaises(CheckerException):
      self.parse(
        """
          /// CHECK-START: Example Group
          /// CHECK-NEXT: bar
        """)

class CheckerParser_SuffixTests(unittest.TestCase):

  noarch_block = """
                  /// CHECK-START: Group
                  /// CHECK:       foo
                  /// CHECK-NEXT:  bar
                  /// CHECK-NOT:   baz
                  /// CHECK-DAG:   yoyo
                """

  arch_block = """
                  /// CHECK-START-{test_arch}: Group
                  /// CHECK:       foo
                  /// CHECK-NEXT:  bar
                  /// CHECK-NOT:   baz
                  /// CHECK-DAG:   yoyo
                """

  def parse(self, checkerText):
    return ParseCheckerStream("<test_file>", "CHECK", io.StringIO(ToUnicode(checkerText)))

  def test_NonArchTests(self):
    for arch in [None] + archs_list:
      checkerFile = self.parse(self.noarch_block)
      self.assertEqual(len(checkerFile.testCases), 1)
      self.assertEqual(len(checkerFile.testCases[0].assertions), 4)

  def test_IgnoreNonTargetArch(self):
    for targetArch in archs_list:
      for testArch in [a for a in archs_list if a != targetArch]:
        checkerText = self.arch_block.format(test_arch = testArch)
        checkerFile = self.parse(checkerText)
        self.assertEqual(len(checkerFile.testCases), 1)
        self.assertEqual(len(checkerFile.testCasesForArch(testArch)), 1)
        self.assertEqual(len(checkerFile.testCasesForArch(targetArch)), 0)

  def test_Arch(self):
    for arch in archs_list:
      checkerText = self.arch_block.format(test_arch = arch)
      checkerFile = self.parse(checkerText)
      self.assertEqual(len(checkerFile.testCases), 1)
      self.assertEqual(len(checkerFile.testCasesForArch(arch)), 1)
      self.assertEqual(len(checkerFile.testCases[0].assertions), 4)

  def test_NoDebugAndArch(self):
    testCase = self.parse("""
        /// CHECK-START: Group
        /// CHECK: foo
        """).testCases[0]
    self.assertFalse(testCase.forDebuggable)
    self.assertEqual(testCase.testArch, None)

  def test_SetDebugNoArch(self):
    testCase = self.parse("""
        /// CHECK-START-DEBUGGABLE: Group
        /// CHECK: foo
        """).testCases[0]
    self.assertTrue(testCase.forDebuggable)
    self.assertEqual(testCase.testArch, None)

  def test_NoDebugSetArch(self):
    testCase = self.parse("""
        /// CHECK-START-ARM: Group
        /// CHECK: foo
        """).testCases[0]
    self.assertFalse(testCase.forDebuggable)
    self.assertEqual(testCase.testArch, "ARM")

  def test_SetDebugAndArch(self):
    testCase = self.parse("""
        /// CHECK-START-ARM-DEBUGGABLE: Group
        /// CHECK: foo
        """).testCases[0]
    self.assertTrue(testCase.forDebuggable)
    self.assertEqual(testCase.testArch, "ARM")

class CheckerParser_EvalTests(unittest.TestCase):
  def parseTestCase(self, string):
    checkerText = u"/// CHECK-START: pass\n" + ToUnicode(string)
    checkerFile = ParseCheckerStream("<test-file>", "CHECK", io.StringIO(checkerText))
    self.assertEqual(len(checkerFile.testCases), 1)
    return checkerFile.testCases[0]

  def parseExpressions(self, string):
    testCase = self.parseTestCase("/// CHECK-EVAL: " + string)
    self.assertEqual(len(testCase.assertions), 1)
    assertion = testCase.assertions[0]
    self.assertEqual(assertion.variant, TestAssertion.Variant.Eval)
    self.assertEqual(assertion.originalText, string)
    return assertion.expressions

  def assertParsesToPlainText(self, text):
    testCase = self.parseTestCase("/// CHECK-EVAL: " + text)
    self.assertEqual(len(testCase.assertions), 1)
    assertion = testCase.assertions[0]
    self.assertEqual(assertion.variant, TestAssertion.Variant.Eval)
    self.assertEqual(assertion.originalText, text)
    self.assertEqual(len(assertion.expressions), 1)
    expression = assertion.expressions[0]
    self.assertEqual(expression.variant, TestExpression.Variant.PlainText)
    self.assertEqual(expression.text, text)

  def test_PlainText(self):
    self.assertParsesToPlainText("XYZ")
    self.assertParsesToPlainText("True")
    self.assertParsesToPlainText("{{abc}}")
    self.assertParsesToPlainText("<<ABC:abc>>")
    self.assertParsesToPlainText("<<ABC=>>")

  def test_VariableReference(self):
    self.assertEqual(self.parseExpressions("<<ABC>>"),
                     [ TestExpression.createVariableReference("ABC") ])
    self.assertEqual(self.parseExpressions("123<<ABC>>"),
                     [ TestExpression.createPlainText("123"),
                       TestExpression.createVariableReference("ABC") ])
    self.assertEqual(self.parseExpressions("123  <<ABC>>"),
                     [ TestExpression.createPlainText("123  "),
                       TestExpression.createVariableReference("ABC") ])
    self.assertEqual(self.parseExpressions("<<ABC>>XYZ"),
                     [ TestExpression.createVariableReference("ABC"),
                       TestExpression.createPlainText("XYZ") ])
    self.assertEqual(self.parseExpressions("<<ABC>>   XYZ"),
                     [ TestExpression.createVariableReference("ABC"),
                       TestExpression.createPlainText("   XYZ") ])
    self.assertEqual(self.parseExpressions("123<<ABC>>XYZ"),
                     [ TestExpression.createPlainText("123"),
                       TestExpression.createVariableReference("ABC"),
                       TestExpression.createPlainText("XYZ") ])
    self.assertEqual(self.parseExpressions("123 <<ABC>>  XYZ"),
                     [ TestExpression.createPlainText("123 "),
                       TestExpression.createVariableReference("ABC"),
                       TestExpression.createPlainText("  XYZ") ])



#!/usr/bin/env python
#
# Copyright (C) 2017 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Outputs quantitative information about Address Sanitizer traces."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import Counter
from datetime import datetime
import argparse
import bisect
import os
import re


def find_match(list_substrings, big_string):
    """Returns the category a trace belongs to by searching substrings."""
    for ind, substr in enumerate(list_substrings):
        if big_string.find(substr) != -1:
            return ind
    return list_substrings.index("Uncategorized")


def absolute_to_relative(data_lists, symbol_traces):
    """Address changed to Dex File offset and shifting time to 0 min in ms."""

    offsets = data_lists["offsets"]
    time_offsets = data_lists["times"]

    # Format of time provided by logcat
    time_format_str = "%H:%M:%S.%f"
    first_access_time = datetime.strptime(data_lists["plot_list"][0][0],
                                          time_format_str)
    for ind, elem in enumerate(data_lists["plot_list"]):
        elem_date_time = datetime.strptime(elem[0], time_format_str)
        # Shift time values so that first access is at time 0 milliseconds
        elem[0] = int((elem_date_time - first_access_time).total_seconds() *
                      1000)
        address_access = int(elem[1], 16)
        # For each poisoned address, find highest Dex File starting address less
        # than address_access
        dex_start_list, dex_size_list = zip(*data_lists["dex_ends_list"])
        dex_file_ind = bisect.bisect(dex_start_list, address_access) - 1
        dex_offset = address_access - dex_start_list[dex_file_ind]
        # Assumes that offsets is already sorted and constrains offset to be
        # within range of the dex_file
        max_offset = min(offsets[1], dex_size_list[dex_file_ind])
        # Meant to nullify data that does not meet offset criteria if specified
        if (dex_offset >= offsets[0] and dex_offset < max_offset and
                elem[0] >= time_offsets[0] and elem[0] < time_offsets[1]):

            elem.insert(1, dex_offset)
            # Category that a data point belongs to
            elem.insert(2, data_lists["cat_list"][ind])
        else:
            elem[:] = 4 * [None]
            symbol_traces[ind] = None
            data_lists["cat_list"][ind] = None


def print_category_info(cat_split, outname, out_dir_name, title):
    """Prints information of category and puts related traces in a files."""
    trace_counts_dict = Counter(cat_split)
    trace_counts_list_ordered = trace_counts_dict.most_common()
    print(53 * "-")
    print(title)
    print("\tNumber of distinct traces: " +
          str(len(trace_counts_list_ordered)))
    print("\tSum of trace counts: " +
          str(sum([trace[1] for trace in trace_counts_list_ordered])))
    print("\n\tCount: How many traces appeared with count\n\t", end="")
    print(Counter([trace[1] for trace in trace_counts_list_ordered]))
    with open(os.path.join(out_dir_name, outname), "w") as output_file:
        for trace in trace_counts_list_ordered:
            output_file.write("\n\nNumber of times appeared: " +
                              str(trace[1]) +
                              "\n")
            output_file.write(trace[0].strip())


def print_categories(categories, symbol_file_split, out_dir_name):
    """Prints details of all categories."""
    symbol_file_split = [trace for trace in symbol_file_split
                         if trace is not None]
    # Info of traces containing a call to current category
    for cat_num, cat_name in enumerate(categories[1:]):
        print("\nCategory #%d" % (cat_num + 1))
        cat_split = [trace for trace in symbol_file_split
                     if cat_name in trace]
        cat_file_name = cat_name.lower() + "cat_output"
        print_category_info(cat_split, cat_file_name, out_dir_name,
                            "Traces containing: " + cat_name)
        noncat_split = [trace for trace in symbol_file_split
                        if cat_name not in trace]
        print_category_info(noncat_split, "non" + cat_file_name,
                            out_dir_name,
                            "Traces not containing: " +
                            cat_name)

    # All traces (including uncategorized) together
    print_category_info(symbol_file_split, "allcat_output",
                        out_dir_name,
                        "All traces together:")
    # Traces containing none of keywords
    # Only used if categories are passed in
    if len(categories) > 1:
        noncat_split = [trace for trace in symbol_file_split if
                        all(cat_name not in trace
                            for cat_name in categories)]
        print_category_info(noncat_split, "noncat_output",
                            out_dir_name,
                            "Uncategorized calls")


def is_directory(path_name):
    """Checks if a path is an actual directory."""
    if not os.path.isdir(path_name):
        dir_error = "%s is not a directory" % (path_name)
        raise argparse.ArgumentTypeError(dir_error)
    return path_name


def parse_args(argv):
    """Parses arguments passed in."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", action="store",
                        default="", dest="out_dir_name", type=is_directory,
                        help="Output Directory")
    parser.add_argument("--dex-file", action="store",
                        default=None, dest="dex_file",
                        type=argparse.FileType("r"),
                        help="Baksmali Dex File Dump")
    parser.add_argument("--offsets", action="store", nargs=2,
                        default=[float(0), float("inf")],
                        dest="offsets",
                        metavar="OFFSET",
                        type=float,
                        help="Filters out accesses not between provided"
                             " offsets if provided. Can provide 'inf'"
                             " for infinity")
    parser.add_argument("--times", action="store", nargs=2,
                        default=[float(0), float("inf")],
                        dest="times",
                        metavar="TIME",
                        type=float,
                        help="Filters out accesses not between provided"
                             " time offsets if provided. Can provide 'inf'"
                             " for infinity")
    parser.add_argument("sanitizer_trace", action="store",
                        type=argparse.FileType("r"),
                        help="File containing sanitizer traces filtered by "
                             "prune_sanitizer_output.py")
    parser.add_argument("symbol_trace", action="store",
                        type=argparse.FileType("r"),
                        help="File containing symbolized traces that match "
                             "sanitizer_trace")
    parser.add_argument("dex_starts", action="store",
                        type=argparse.FileType("r"),
                        help="File containing starting addresses of Dex Files")
    parser.add_argument("categories", action="store", nargs="*",
                        help="Keywords expected to show in large amounts of"
                             " symbolized traces")

    return parser.parse_args(argv)


def get_dex_offset_data(line, dex_file_item):
    """ Returns a tuple of dex file offset, item name, and data of a line."""
    return (int(line[:line.find(":")], 16),
            (dex_file_item,
             line.split("|")[1].strip())
            )


def read_data(parsed_argv):
    """Reads data from filepath arguments and parses them into lists."""
    # Using a dictionary to establish relation between lists added
    data_lists = {}
    categories = parsed_argv.categories
    # Makes sure each trace maps to some category
    categories.insert(0, "Uncategorized")

    data_lists["offsets"] = parsed_argv.offsets
    data_lists["offsets"].sort()

    data_lists["times"] = parsed_argv.times
    data_lists["times"].sort()

    logcat_file_data = parsed_argv.sanitizer_trace.readlines()
    parsed_argv.sanitizer_trace.close()

    symbol_file_split = parsed_argv.symbol_trace.read().split("Stack Trace")
    # Removes text before first trace
    symbol_file_split = symbol_file_split[1:]
    parsed_argv.symbol_trace.close()

    dex_start_file_data = parsed_argv.dex_starts.readlines()
    parsed_argv.dex_starts.close()

    if parsed_argv.dex_file is not None:
        dex_file_data = parsed_argv.dex_file.read()
        parsed_argv.dex_file.close()
        # Splits baksmali dump by each item
        item_split = [s.splitlines() for s in re.split(r"\|\[[0-9]+\] ",
                                                       dex_file_data)]
        # Splits each item by line and creates a list of offsets and a
        # corresponding list of the data associated with that line
        offset_list, offset_data = zip(*[get_dex_offset_data(line, item[0])
                                         for item in item_split
                                         for line in item[1:]
                                         if re.search("[0-9a-f]{6}:", line)
                                         is not None and
                                         line.find("|") != -1])
        data_lists["offset_list"] = offset_list
        data_lists["offset_data"] = offset_data
    else:
        dex_file_data = None

    # Each element is a tuple of time and address accessed
    data_lists["plot_list"] = [[elem[1] for elem in enumerate(line.split())
                                if elem[0] in (1, 11)
                                ]
                               for line in logcat_file_data
                               if "use-after-poison" in line or
                               "unknown-crash" in line
                               ]
    # Contains a mapping between traces and the category they belong to
    # based on arguments
    data_lists["cat_list"] = [categories[find_match(categories, trace)]
                              for trace in symbol_file_split]

    # Contains a list of starting address of all dex files to calculate dex
    # offsets
    data_lists["dex_ends_list"] = [(int(line.split()[9], 16),
                                    int(line.split()[12])
                                    )
                                   for line in dex_start_file_data
                                   if "RegisterDexFile" in line
                                   ]
    # Dex File Starting addresses must be sorted because bisect requires sorted
    # lists.
    data_lists["dex_ends_list"].sort()

    return data_lists, categories, symbol_file_split


def main():
    """Takes in trace information and outputs details about them."""
    parsed_argv = parse_args(None)
    data_lists, categories, symbol_file_split = read_data(parsed_argv)

    # Formats plot_list such that each element is a data point
    absolute_to_relative(data_lists, symbol_file_split)
    for file_ext, cat_name in enumerate(categories):
        out_file_name = os.path.join(parsed_argv.out_dir_name, "time_output_" +
                                     str(file_ext) +
                                     ".dat")
        with open(out_file_name, "w") as output_file:
            output_file.write("# Category: " + cat_name + "\n")
            output_file.write("# Time, Dex File Offset_10, Dex File Offset_16,"
                              " Address, Item Accessed, Item Member Accessed"
                              " Unaligned\n")
            for time, dex_offset, category, address in data_lists["plot_list"]:
                if category == cat_name:
                    output_file.write(
                        str(time) +
                        " " +
                        str(dex_offset) +
                        " #" +
                        hex(dex_offset) +
                        " " +
                        str(address))
                    if "offset_list" in data_lists:
                        dex_offset_index = bisect.bisect(
                            data_lists["offset_list"],
                            dex_offset) - 1
                        aligned_dex_offset = (data_lists["offset_list"]
                                                        [dex_offset_index])
                        dex_offset_data = (data_lists["offset_data"]
                                                     [dex_offset_index])
                        output_file.write(
                            " " +
                            "|".join(dex_offset_data) +
                            " " +
                            str(aligned_dex_offset != dex_offset))
                    output_file.write("\n")
    print_categories(categories, symbol_file_split, parsed_argv.out_dir_name)


if __name__ == "__main__":
    main()



#!/usr/bin/env python2
#
# Copyright (C) 2014 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os

from common.archs                     import archs_list
from common.logger                    import Logger
from file_format.c1visualizer.parser  import ParseC1visualizerStream
from file_format.checker.parser       import ParseCheckerStream
from match.file                       import MatchFiles

def ParseArguments():
  parser = argparse.ArgumentParser()
  parser.add_argument("tested_file",
                      help="text file the checks should be verified against")
  parser.add_argument("source_path", nargs="?",
                      help="path to file/folder with checking annotations")
  parser.add_argument("--check-prefix", dest="check_prefix", default="CHECK", metavar="PREFIX",
                      help="prefix of checks in the test files (default: CHECK)")
  parser.add_argument("--list-passes", dest="list_passes", action="store_true",
                      help="print a list of all passes found in the tested file")
  parser.add_argument("--dump-pass", dest="dump_pass", metavar="PASS",
                      help="print a compiler pass dump")
  parser.add_argument("--arch", dest="arch", choices=archs_list,
                      help="Run tests for the specified target architecture.")
  parser.add_argument("--debuggable", action="store_true",
                      help="Run tests for debuggable code.")
  parser.add_argument("-q", "--quiet", action="store_true",
                      help="print only errors")
  return parser.parse_args()


def ListPasses(outputFilename):
  c1File = ParseC1visualizerStream(os.path.basename(outputFilename), open(outputFilename, "r"))
  for compiler_pass in c1File.passes:
    Logger.log(compiler_pass.name)


def DumpPass(outputFilename, passName):
  c1File = ParseC1visualizerStream(os.path.basename(outputFilename), open(outputFilename, "r"))
  compiler_pass = c1File.findPass(passName)
  if compiler_pass:
    maxLineNo = compiler_pass.startLineNo + len(compiler_pass.body)
    lenLineNo = len(str(maxLineNo)) + 2
    curLineNo = compiler_pass.startLineNo
    for line in compiler_pass.body:
      Logger.log((str(curLineNo) + ":").ljust(lenLineNo) + line)
      curLineNo += 1
  else:
    Logger.fail("Pass \"" + passName + "\" not found in the output")


def FindCheckerFiles(path):
  """ Returns a list of files to scan for check annotations in the given path.
      Path to a file is returned as a single-element list, directories are
      recursively traversed and all '.java' and '.smali' files returned.
  """
  if not path:
    Logger.fail("No source path provided")
  elif os.path.isfile(path):
    return [ path ]
  elif os.path.isdir(path):
    foundFiles = []
    for root, dirs, files in os.walk(path):
      for file in files:
        extension = os.path.splitext(file)[1]
        if extension in [".java", ".smali"]:
          foundFiles.append(os.path.join(root, file))
    return foundFiles
  else:
    Logger.fail("Source path \"" + path + "\" not found")


def RunTests(checkPrefix, checkPath, outputFilename, targetArch, debuggableMode):
  c1File = ParseC1visualizerStream(os.path.basename(outputFilename), open(outputFilename, "r"))
  for checkFilename in FindCheckerFiles(checkPath):
    checkerFile = ParseCheckerStream(os.path.basename(checkFilename),
                                     checkPrefix,
                                     open(checkFilename, "r"),
                                     targetArch)
    MatchFiles(checkerFile, c1File, targetArch, debuggableMode)


if __name__ == "__main__":
  args = ParseArguments()

  if args.quiet:
    Logger.Verbosity = Logger.Level.Error

  if args.list_passes:
    ListPasses(args.tested_file)
  elif args.dump_pass:
    DumpPass(args.tested_file, args.dump_pass)
  else:
    RunTests(args.check_prefix, args.source_path, args.tested_file, args.arch, args.debuggable)



#!/usr/bin/env python3
#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing common logic from python testing tools."""

import abc
import os
import signal
import shlex
import shutil
import time

from enum import Enum
from enum import unique

from subprocess import DEVNULL
from subprocess import check_call
from subprocess import PIPE
from subprocess import Popen
from subprocess import STDOUT
from subprocess import TimeoutExpired

from tempfile import mkdtemp
from tempfile import NamedTemporaryFile

# Temporary directory path on device.
DEVICE_TMP_PATH = '/data/local/tmp'

# Architectures supported in dalvik cache.
DALVIK_CACHE_ARCHS = ['arm', 'arm64', 'x86', 'x86_64']


@unique
class RetCode(Enum):
  """Enum representing normalized return codes."""
  SUCCESS = 0
  TIMEOUT = 1
  ERROR = 2
  NOTCOMPILED = 3
  NOTRUN = 4


@unique
class LogSeverity(Enum):
  VERBOSE = 0
  DEBUG = 1
  INFO = 2
  WARNING = 3
  ERROR = 4
  FATAL = 5
  SILENT = 6

  @property
  def symbol(self):
    return self.name[0]

  @classmethod
  def FromSymbol(cls, s):
    for log_severity in LogSeverity:
      if log_severity.symbol == s:
        return log_severity
    raise ValueError("{0} is not a valid log severity symbol".format(s))

  def __ge__(self, other):
    if self.__class__ is other.__class__:
      return self.value >= other.value
    return NotImplemented

  def __gt__(self, other):
    if self.__class__ is other.__class__:
      return self.value > other.value
    return NotImplemented

  def __le__(self, other):
    if self.__class__ is other.__class__:
      return self.value <= other.value
    return NotImplemented

  def __lt__(self, other):
    if self.__class__ is other.__class__:
      return self.value < other.value
    return NotImplemented


def GetEnvVariableOrError(variable_name):
  """Gets value of an environmental variable.

  If the variable is not set raises FatalError.

  Args:
    variable_name: string, name of variable to get.

  Returns:
    string, value of requested variable.

  Raises:
    FatalError: Requested variable is not set.
  """
  top = os.environ.get(variable_name)
  if top is None:
    raise FatalError('{0} environmental variable not set.'.format(
        variable_name))
  return top


def _DexArchCachePaths(android_data_path):
  """Returns paths to architecture specific caches.

  Args:
    android_data_path: string, path dalvik-cache resides in.

  Returns:
    Iterable paths to architecture specific caches.
  """
  return ('{0}/dalvik-cache/{1}'.format(android_data_path, arch)
          for arch in DALVIK_CACHE_ARCHS)


def RunCommandForOutput(cmd, env, stdout, stderr, timeout=60):
  """Runs command piping output to files, stderr or stdout.

  Args:
    cmd: list of strings, command to run.
    env: shell environment to run the command with.
    stdout: file handle or one of Subprocess.PIPE, Subprocess.STDOUT,
      Subprocess.DEVNULL, see Popen.
    stderr: file handle or one of Subprocess.PIPE, Subprocess.STDOUT,
      Subprocess.DEVNULL, see Popen.
    timeout: int, timeout in seconds.

  Returns:
    tuple (string, string, RetCode) stdout output, stderr output, normalized
      return code.
  """
  proc = Popen(cmd, stdout=stdout, stderr=stderr, env=env,
               universal_newlines=True, start_new_session=True)
  try:
    (output, stderr_output) = proc.communicate(timeout=timeout)
    if proc.returncode == 0:
      retcode = RetCode.SUCCESS
    else:
      retcode = RetCode.ERROR
  except TimeoutExpired:
    os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
    (output, stderr_output) = proc.communicate()
    retcode = RetCode.TIMEOUT
  return (output, stderr_output, retcode)


def _LogCmdOutput(logfile, cmd, output, retcode):
  """Logs output of a command.

  Args:
    logfile: file handle to logfile.
    cmd: list of strings, command.
    output: command output.
    retcode: RetCode, normalized retcode.
  """
  logfile.write('Command:\n{0}\n{1}\nReturn code: {2}\n'.format(
      CommandListToCommandString(cmd), output, retcode))


def RunCommand(cmd, out, err, timeout=5):
  """Executes a command, and returns its return code.

  Args:
    cmd: list of strings, a command to execute
    out: string, file name to open for stdout (or None)
    err: string, file name to open for stderr (or None)
    timeout: int, time out in seconds
  Returns:
    RetCode, return code of running command (forced RetCode.TIMEOUT
    on timeout)
  """
  devnull = DEVNULL
  outf = devnull
  if out is not None:
    outf = open(out, mode='w')
  errf = devnull
  if err is not None:
    errf = open(err, mode='w')
  (_, _, retcode) = RunCommandForOutput(cmd, None, outf, errf, timeout)
  if outf != devnull:
    outf.close()
  if errf != devnull:
    errf.close()
  return retcode


def CommandListToCommandString(cmd):
  """Converts shell command represented as list of strings to a single string.

  Each element of the list is wrapped in double quotes.

  Args:
    cmd: list of strings, shell command.

  Returns:
    string, shell command.
  """
  return ' '.join([shlex.quote(segment) for segment in cmd])


class FatalError(Exception):
  """Fatal error in script."""


class ITestEnv(object):
  """Test environment abstraction.

  Provides unified interface for interacting with host and device test
  environments. Creates a test directory and expose methods to modify test files
  and run commands.
  """
  __meta_class__ = abc.ABCMeta

  @abc.abstractmethod
  def CreateFile(self, name=None):
    """Creates a file in test directory.

    Returned path to file can be used in commands run in the environment.

    Args:
      name: string, file name. If None file is named arbitrarily.

    Returns:
      string, environment specific path to file.
    """

  @abc.abstractmethod
  def WriteLines(self, file_path, lines):
    """Writes lines to a file in test directory.

    If file exists it gets overwritten. If file doest not exist it is created.

    Args:
      file_path: string, environment specific path to file.
      lines: list of strings to write.
    """

  @abc.abstractmethod
  def RunCommand(self, cmd, log_severity=LogSeverity.ERROR):
    """Runs command in environment.

    Args:
      cmd: list of strings, command to run.
      log_severity: LogSeverity, minimum severity of logs included in output.
    Returns:
      tuple (string, int) output, return code.
    """

  @abc.abstractproperty
  def logfile(self):
    """Gets file handle to logfile residing on host."""


class HostTestEnv(ITestEnv):
  """Host test environment. Concrete implementation of ITestEnv.

  Maintains a test directory in /tmp/. Runs commands on the host in modified
  shell environment. Mimics art script behavior.

  For methods documentation see base class.
  """

  def __init__(self, directory_prefix, cleanup=True, logfile_path=None,
               timeout=60, x64=False):
    """Constructor.

    Args:
      directory_prefix: string, prefix for environment directory name.
      cleanup: boolean, if True remove test directory in destructor.
      logfile_path: string, can be used to specify custom logfile location.
      timeout: int, seconds, time to wait for single test run to finish.
      x64: boolean, whether to setup in x64 mode.
    """
    self._cleanup = cleanup
    self._timeout = timeout
    self._env_path = mkdtemp(dir='/tmp/', prefix=directory_prefix)
    if logfile_path is None:
      self._logfile = open('{0}/log'.format(self._env_path), 'w+')
    else:
      self._logfile = open(logfile_path, 'w+')
    os.mkdir('{0}/dalvik-cache'.format(self._env_path))
    for arch_cache_path in _DexArchCachePaths(self._env_path):
      os.mkdir(arch_cache_path)
    lib = 'lib64' if x64 else 'lib'
    android_root = GetEnvVariableOrError('ANDROID_HOST_OUT')
    android_i18n_root = android_root + '/com.android.i18n'
    android_art_root = android_root + '/com.android.art'
    android_tzdata_root = android_root + '/com.android.tzdata'
    library_path = android_root + '/' + lib
    path = android_root + '/bin'
    self._shell_env = os.environ.copy()
    self._shell_env['ANDROID_DATA'] = self._env_path
    self._shell_env['ANDROID_ROOT'] = android_root
    self._shell_env['ANDROID_I18N_ROOT'] = android_i18n_root
    self._shell_env['ANDROID_ART_ROOT'] = android_art_root
    self._shell_env['ANDROID_TZDATA_ROOT'] = android_tzdata_root
    self._shell_env['LD_LIBRARY_PATH'] = library_path
    self._shell_env['DYLD_LIBRARY_PATH'] = library_path
    self._shell_env['PATH'] = (path + ':' + self._shell_env['PATH'])
    # Using dlopen requires load bias on the host.
    self._shell_env['LD_USE_LOAD_BIAS'] = '1'

  def __del__(self):
    if self._cleanup:
      shutil.rmtree(self._env_path)

  def CreateFile(self, name=None):
    if name is None:
      f = NamedTemporaryFile(dir=self._env_path, delete=False)
    else:
      f = open('{0}/{1}'.format(self._env_path, name), 'w+')
    return f.name

  def WriteLines(self, file_path, lines):
    with open(file_path, 'w') as f:
      f.writelines('{0}\n'.format(line) for line in lines)
    return

  def RunCommand(self, cmd, log_severity=LogSeverity.ERROR):
    self._EmptyDexCache()
    env = self._shell_env.copy()
    env.update({'ANDROID_LOG_TAGS':'*:' + log_severity.symbol.lower()})
    (output, err_output, retcode) = RunCommandForOutput(
        cmd, env, PIPE, PIPE, self._timeout)
    # We append err_output to output to stay consistent with DeviceTestEnv
    # implementation.
    output += err_output
    _LogCmdOutput(self._logfile, cmd, output, retcode)
    return (output, retcode)

  @property
  def logfile(self):
    return self._logfile

  def _EmptyDexCache(self):
    """Empties dex cache.

    Iterate over files in architecture specific cache directories and remove
    them.
    """
    for arch_cache_path in _DexArchCachePaths(self._env_path):
      for file_path in os.listdir(arch_cache_path):
        file_path = '{0}/{1}'.format(arch_cache_path, file_path)
        if os.path.isfile(file_path):
          os.unlink(file_path)


class DeviceTestEnv(ITestEnv):
  """Device test environment. Concrete implementation of ITestEnv.

  For methods documentation see base class.
  """

  def __init__(self, directory_prefix, cleanup=True, logfile_path=None,
               timeout=60, specific_device=None):
    """Constructor.

    Args:
      directory_prefix: string, prefix for environment directory name.
      cleanup: boolean, if True remove test directory in destructor.
      logfile_path: string, can be used to specify custom logfile location.
      timeout: int, seconds, time to wait for single test run to finish.
      specific_device: string, serial number of device to use.
    """
    self._cleanup = cleanup
    self._timeout = timeout
    self._specific_device = specific_device
    self._host_env_path = mkdtemp(dir='/tmp/', prefix=directory_prefix)
    if logfile_path is None:
      self._logfile = open('{0}/log'.format(self._host_env_path), 'w+')
    else:
      self._logfile = open(logfile_path, 'w+')
    self._device_env_path = '{0}/{1}'.format(
        DEVICE_TMP_PATH, os.path.basename(self._host_env_path))
    self._shell_env = os.environ.copy()

    self._AdbMkdir('{0}/dalvik-cache'.format(self._device_env_path))
    for arch_cache_path in _DexArchCachePaths(self._device_env_path):
      self._AdbMkdir(arch_cache_path)

  def __del__(self):
    if self._cleanup:
      shutil.rmtree(self._host_env_path)
      check_call(shlex.split(
          'adb shell if [ -d "{0}" ]; then rm -rf "{0}"; fi'
          .format(self._device_env_path)))

  def CreateFile(self, name=None):
    with NamedTemporaryFile(mode='w') as temp_file:
      self._AdbPush(temp_file.name, self._device_env_path)
      if name is None:
        name = os.path.basename(temp_file.name)
      return '{0}/{1}'.format(self._device_env_path, name)

  def WriteLines(self, file_path, lines):
    with NamedTemporaryFile(mode='w') as temp_file:
      temp_file.writelines('{0}\n'.format(line) for line in lines)
      temp_file.flush()
      self._AdbPush(temp_file.name, file_path)
    return

  def _ExtractPid(self, brief_log_line):
    """Extracts PID from a single logcat line in brief format."""
    pid_start_idx = brief_log_line.find('(') + 2
    if pid_start_idx == -1:
      return None
    pid_end_idx = brief_log_line.find(')', pid_start_idx)
    if pid_end_idx == -1:
      return None
    return brief_log_line[pid_start_idx:pid_end_idx]

  def _ExtractSeverity(self, brief_log_line):
    """Extracts LogSeverity from a single logcat line in brief format."""
    if not brief_log_line:
      return None
    return LogSeverity.FromSymbol(brief_log_line[0])

  def RunCommand(self, cmd, log_severity=LogSeverity.ERROR):
    self._EmptyDexCache()
    env_vars_cmd = 'ANDROID_DATA={0} ANDROID_LOG_TAGS=*:i'.format(
        self._device_env_path)
    adb_cmd = ['adb']
    if self._specific_device:
      adb_cmd += ['-s', self._specific_device]
    logcat_cmd = adb_cmd + ['logcat', '-v', 'brief', '-s', '-b', 'main',
                            '-T', '1', 'dex2oat:*', 'dex2oatd:*']
    logcat_proc = Popen(logcat_cmd, stdout=PIPE, stderr=STDOUT,
                        universal_newlines=True)
    cmd_str = CommandListToCommandString(cmd)
    # Print PID of the shell and exec command. We later retrieve this PID and
    # use it to filter dex2oat logs, keeping those with matching parent PID.
    device_cmd = ('echo $$ && ' + env_vars_cmd + ' exec ' + cmd_str)
    cmd = adb_cmd + ['shell', device_cmd]
    (output, _, retcode) = RunCommandForOutput(cmd, self._shell_env, PIPE,
                                               STDOUT, self._timeout)
    # We need to make sure to only kill logcat once all relevant logs arrive.
    # Sleep is used for simplicity.
    time.sleep(0.5)
    logcat_proc.kill()
    end_of_first_line = output.find('\n')
    if end_of_first_line != -1:
      parent_pid = output[:end_of_first_line]
      output = output[end_of_first_line + 1:]
      logcat_output, _ = logcat_proc.communicate()
      logcat_lines = logcat_output.splitlines(keepends=True)
      dex2oat_pids = []
      for line in logcat_lines:
        # Dex2oat was started by our runtime instance.
        if 'Running dex2oat (parent PID = ' + parent_pid in line:
          dex2oat_pids.append(self._ExtractPid(line))
          break
      if dex2oat_pids:
        for line in logcat_lines:
          if (self._ExtractPid(line) in dex2oat_pids and
              self._ExtractSeverity(line) >= log_severity):
            output += line
    _LogCmdOutput(self._logfile, cmd, output, retcode)
    return (output, retcode)

  @property
  def logfile(self):
    return self._logfile

  def PushClasspath(self, classpath):
    """Push classpath to on-device test directory.

    Classpath can contain multiple colon separated file paths, each file is
    pushed. Returns analogous classpath with paths valid on device.

    Args:
      classpath: string, classpath in format 'a/b/c:d/e/f'.
    Returns:
      string, classpath valid on device.
    """
    paths = classpath.split(':')
    device_paths = []
    for path in paths:
      device_paths.append('{0}/{1}'.format(
          self._device_env_path, os.path.basename(path)))
      self._AdbPush(path, self._device_env_path)
    return ':'.join(device_paths)

  def _AdbPush(self, what, where):
    check_call(shlex.split('adb push "{0}" "{1}"'.format(what, where)),
               stdout=self._logfile, stderr=self._logfile)

  def _AdbMkdir(self, path):
    check_call(shlex.split('adb shell mkdir "{0}" -p'.format(path)),
               stdout=self._logfile, stderr=self._logfile)

  def _EmptyDexCache(self):
    """Empties dex cache."""
    for arch_cache_path in _DexArchCachePaths(self._device_env_path):
      cmd = 'adb shell if [ -d "{0}" ]; then rm -f "{0}"/*; fi'.format(
          arch_cache_path)
      check_call(shlex.split(cmd), stdout=self._logfile, stderr=self._logfile)



# Copyright (C) 2014 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def SplitStream(stream, fnProcessLine, fnLineOutsideChunk):
  """ Reads the given input stream and splits it into chunks based on
      information extracted from individual lines.

  Arguments:
   - fnProcessLine: Called on each line with the text and line number. Must
     return a triplet, composed of the name of the chunk started on this line,
     the data extracted, and the name of the architecture this test applies to
     (or None to indicate that all architectures should run this test).
   - fnLineOutsideChunk: Called on attempt to attach data prior to creating
     a chunk.
  """
  lineNo = 0
  allChunks = []
  currentChunk = None

  for line in stream:
    lineNo += 1
    line = line.strip()
    if not line:
      continue

    # Let the child class process the line and return information about it.
    # The _processLine method can modify the content of the line (or delete it
    # entirely) and specify whether it starts a new group.
    processedLine, newChunkName, testArch = fnProcessLine(line, lineNo)
    # Currently, only a full chunk can be specified as architecture-specific.
    assert testArch is None or newChunkName is not None
    if newChunkName is not None:
      currentChunk = (newChunkName, [], lineNo, testArch)
      allChunks.append(currentChunk)
    if processedLine is not None:
      if currentChunk is not None:
        currentChunk[1].append(processedLine)
      else:
        fnLineOutsideChunk(line, lineNo)
  return allChunks



#!/usr/bin/python3
#
# Copyright (C) 2015 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generate java test files for test 966.
"""

import generate_smali as base
import os
import sys
from pathlib import Path

BUILD_TOP = os.getenv("ANDROID_BUILD_TOP")
if BUILD_TOP is None:
  print("ANDROID_BUILD_TOP not set. Please run build/envsetup.sh", file=sys.stderr)
  sys.exit(1)

# Allow us to import mixins.
sys.path.append(str(Path(BUILD_TOP)/"art"/"test"/"utils"/"python"))

import testgen.mixins as mixins
import functools
import operator
import subprocess

class JavaConverter(mixins.DumpMixin, mixins.Named, mixins.JavaFileMixin):
  """
  A class that can convert a SmaliFile to a JavaFile.
  """
  def __init__(self, inner):
    self.inner = inner

  def get_name(self):
    """Gets the name of this file."""
    return self.inner.get_name()

  def __str__(self):
    out = ""
    for line in str(self.inner).splitlines(keepends = True):
      if line.startswith("#"):
        out += line[1:]
    return out

class Compiler:
  def __init__(self, sources, javac, temp_dir, classes_dir):
    self.javac = javac
    self.temp_dir = temp_dir
    self.classes_dir = classes_dir
    self.sources = sources

  def compile_files(self, args, files):
    """
    Compile the files given with the arguments given.
    """
    args = args.split()
    files = list(map(str, files))
    cmd = ['sh', '-a', '-e', '--', str(self.javac)] + args + files
    print("Running compile command: {}".format(cmd))
    subprocess.check_call(cmd)
    print("Compiled {} files".format(len(files)))

  def execute(self):
    """
    Compiles this test, doing partial compilation as necessary.
    """
    # Compile Main and all classes first. Force all interfaces to be default so that there will be
    # no compiler problems (works since classes only implement 1 interface).
    for f in self.sources:
      if isinstance(f, base.TestInterface):
        JavaConverter(f.get_specific_version(base.InterfaceType.default)).dump(self.temp_dir)
      else:
        JavaConverter(f).dump(self.temp_dir)
    self.compile_files("-d {}".format(self.classes_dir), self.temp_dir.glob("*.java"))

    # Now we compile the interfaces
    ifaces = set(i for i in self.sources if isinstance(i, base.TestInterface))
    filters = (lambda a: a.is_default(), lambda a: not a.is_default())
    converters = (lambda a: JavaConverter(a.get_specific_version(base.InterfaceType.default)),
                  lambda a: JavaConverter(a.get_specific_version(base.InterfaceType.empty)))
    while len(ifaces) != 0:
      for iface_filter, iface_converter in zip(filters, converters):
        # Find those ifaces where there are no (uncompiled) interfaces that are subtypes.
        tops = set(filter(lambda a: iface_filter(a) and not any(map(lambda i: a in i.get_super_types(), ifaces)), ifaces))
        files = []
        # Dump these ones, they are getting compiled.
        for f in tops:
          out = JavaConverter(f)
          out.dump(self.temp_dir)
          files.append(self.temp_dir / out.get_file_name())
        # Force all superinterfaces of these to be empty so there will be no conflicts
        overrides = functools.reduce(operator.or_, map(lambda i: i.get_super_types(), tops), set())
        for overridden in overrides:
          out = iface_converter(overridden)
          out.dump(self.temp_dir)
          files.append(self.temp_dir / out.get_file_name())
        self.compile_files("-d {outdir} -cp {outdir}".format(outdir = self.classes_dir), files)
        # Remove these from the set of interfaces to be compiled.
        ifaces -= tops
    print("Finished compiling all files.")
    return

def main(argv):
  javac_exec = Path(argv[1])
  if not javac_exec.exists() or not javac_exec.is_file():
    print("{} is not a shell script".format(javac_exec), file=sys.stderr)
    sys.exit(1)
  temp_dir = Path(argv[2])
  if not temp_dir.exists() or not temp_dir.is_dir():
    print("{} is not a valid source dir".format(temp_dir), file=sys.stderr)
    sys.exit(1)
  classes_dir = Path(argv[3])
  if not classes_dir.exists() or not classes_dir.is_dir():
    print("{} is not a valid classes directory".format(classes_dir), file=sys.stderr)
    sys.exit(1)
  expected_txt = Path(argv[4])
  mainclass, all_files = base.create_all_test_files()

  with expected_txt.open('w') as out:
    print(mainclass.get_expected(), file=out)
  print("Wrote expected output")

  Compiler(all_files, javac_exec, temp_dir, classes_dir).execute()

if __name__ == '__main__':
  main(sys.argv)



#!/usr/bin/python3
#
# Copyright (C) 2015 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common functions useful for writing test generators in python
"""

import itertools
import os
import string
from pathlib import Path

BUILD_TOP = os.getenv("ANDROID_BUILD_TOP")
if BUILD_TOP is None:
  print("ANDROID_BUILD_TOP not set. Please run build/envsetup.sh", file=sys.stderr)
  sys.exit(1)

# An iterator which yields strings made from lowercase letters. First yields
# all 1 length strings, then all 2 and so on. It does this alphabetically.
NAME_GEN = itertools.chain.from_iterable(
    map(lambda n: itertools.product(string.ascii_lowercase, repeat=n),
        itertools.count(1)))

def gensym():
  """
  Returns a new, globally unique, identifier name that is a valid Java symbol
  on each call.
  """
  return ''.join(next(NAME_GEN))

def filter_blanks(s):
  """
  Takes a string returns the same string sans empty lines
  """
  return "\n".join(a for a in s.split("\n") if a.strip() != "")

def get_copyright(filetype = "java"):
  """
  Returns the standard copyright header for the given filetype
  """
  if filetype == "smali":
    return "\n".join(map(lambda a: "# " + a, get_copyright("java").split("\n")))
  else:
    fname = filetype + ".txt"
    with (Path(BUILD_TOP)/"development"/"docs"/"copyright-templates"/fname).open() as template:
      return "".join(template.readlines())

def subtree_sizes(n):
  """
  A generator that yields a tuple containing a possible arrangement of subtree
  nodes for a tree with a total of 'n' leaf nodes.
  """
  if n == 0:
    return
  elif n == 1:
    yield (0,)
  elif n == 2:
    yield (1, 1)
  else:
    for prevt in subtree_sizes(n - 1):
      prev = list(prevt)
      yield tuple([1] + prev)
      for i in range(len(prev)):
        prev[i] += 1
        yield tuple(prev)
        prev[i] -= 1




#!/usr/bin/env python
#
# Copyright (C) 2014 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script that parses a trace filed produced in streaming mode. The file is broken up into
   a header and body part, which, when concatenated, make up a non-streaming trace file that
   can be used with traceview."""

import sys

class MyException(Exception):
  pass

class BufferUnderrun(Exception):
  pass

def ReadShortLE(f):
  byte1 = f.read(1)
  if not byte1:
    raise BufferUnderrun()
  byte2 = f.read(1)
  if not byte2:
    raise BufferUnderrun()
  return ord(byte1) + (ord(byte2) << 8);

def WriteShortLE(f, val):
  bytes = [ (val & 0xFF), ((val >> 8) & 0xFF) ]
  asbytearray = bytearray(bytes)
  f.write(asbytearray)

def ReadIntLE(f):
  byte1 = f.read(1)
  if not byte1:
    raise BufferUnderrun()
  byte2 = f.read(1)
  if not byte2:
    raise BufferUnderrun()
  byte3 = f.read(1)
  if not byte3:
    raise BufferUnderrun()
  byte4 = f.read(1)
  if not byte4:
    raise BufferUnderrun()
  return ord(byte1) + (ord(byte2) << 8) + (ord(byte3) << 16) + (ord(byte4) << 24);

def WriteIntLE(f, val):
  bytes = [ (val & 0xFF), ((val >> 8) & 0xFF), ((val >> 16) & 0xFF), ((val >> 24) & 0xFF) ]
  asbytearray = bytearray(bytes)
  f.write(asbytearray)

def Copy(input, output, length):
  buf = input.read(length)
  if len(buf) != length:
    raise BufferUnderrun()
  output.write(buf)

class Rewriter:

  def PrintHeader(self, header):
    header.write('*version\n');
    header.write('3\n');
    header.write('data-file-overflow=false\n');
    header.write('clock=dual\n');
    header.write('vm=art\n');

  def ProcessDataHeader(self, input, body):
    magic = ReadIntLE(input)
    if magic != 0x574f4c53:
      raise MyException("Magic wrong")

    WriteIntLE(body, magic)

    version = ReadShortLE(input)
    if (version & 0xf0) != 0xf0:
      raise MyException("Does not seem to be a streaming trace: %d." % version)
    version = version ^ 0xf0

    if version != 3:
      raise MyException("Only support version 3")

    WriteShortLE(body, version)

    # read offset
    offsetToData = ReadShortLE(input) - 16
    WriteShortLE(body, offsetToData + 16)

    # copy startWhen
    Copy(input, body, 8)

    if version == 1:
      self._mRecordSize = 9;
    elif version == 2:
      self._mRecordSize = 10;
    else:
      self._mRecordSize = ReadShortLE(input)
      WriteShortLE(body, self._mRecordSize)
      offsetToData -= 2;

    # Skip over offsetToData bytes
    Copy(input, body, offsetToData)

  def ProcessMethod(self, input):
    stringLength = ReadShortLE(input)
    str = input.read(stringLength)
    self._methods.append(str)
    print 'New method: %s' % str

  def ProcessThread(self, input):
    tid = ReadShortLE(input)
    stringLength = ReadShortLE(input)
    str = input.read(stringLength)
    self._threads.append('%d\t%s\n' % (tid, str))
    print 'New thread: %d/%s' % (tid, str)

  def ProcessTraceSummary(self, input):
    summaryLength = ReadIntLE(input)
    str = input.read(summaryLength)
    self._summary = str
    print 'Summary: \"%s\"' % str

  def ProcessSpecial(self, input):
    code = ord(input.read(1))
    if code == 1:
      self.ProcessMethod(input)
    elif code == 2:
      self.ProcessThread(input)
    elif code == 3:
      self.ProcessTraceSummary(input)
    else:
      raise MyException("Unknown special!")

  def Process(self, input, body):
    try:
      while True:
        threadId = ReadShortLE(input)
        if threadId == 0:
          self.ProcessSpecial(input)
        else:
          # Regular package, just copy
          WriteShortLE(body, threadId)
          Copy(input, body, self._mRecordSize - 2)
    except BufferUnderrun:
      print 'Buffer underrun, file was probably truncated. Results should still be usable.'

  def Finalize(self, header):
    # If the summary is present in the input file, use it as the header except
    # for the methods section which is emtpy in the input file. If not present,
    # apppend header with the threads that are recorded in the input stream.
    if (self._summary):
      # Erase the contents that's already written earlier by PrintHeader.
      header.seek(0)
      header.truncate()
      # Copy the lines from the input summary to the output header until
      # the methods section is seen.
      for line in self._summary.splitlines(True):
        if line == "*methods\n":
          break
        else:
          header.write(line)
    else:
      header.write('*threads\n')
      for t in self._threads:
        header.write(t)
    header.write('*methods\n')
    for m in self._methods:
      header.write(m)
    header.write('*end\n')

  def ProcessFile(self, filename):
    input = open(filename, 'rb')                     # Input file
    header = open(filename + '.header', 'w')         # Header part
    body = open(filename + '.body', 'wb')            # Body part

    self.PrintHeader(header)

    self.ProcessDataHeader(input, body)

    self._methods = []
    self._threads = []
    self._summary = None
    self.Process(input, body)

    self.Finalize(header)

    input.close()
    header.close()
    body.close()

def main():
  Rewriter().ProcessFile(sys.argv[1])
  header_name = sys.argv[1] + '.header'
  body_name = sys.argv[1] + '.body'
  print 'Results have been written to %s and %s.' % (header_name, body_name)
  print 'Concatenate the files to get a result usable with traceview.'
  sys.exit(0)

if __name__ == '__main__':
  main()


#!/usr/bin/env python3
#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import shutil
import sys

from glob import glob
from subprocess import call
from tempfile import mkdtemp

sys.path.append(os.path.dirname(os.path.dirname(
        os.path.realpath(__file__))))

from common.common import FatalError
from common.common import GetEnvVariableOrError
from common.common import RetCode
from common.common import RunCommand


#
# Tester class.
#


class DexFuzzTester(object):
  """Tester that feeds JFuzz programs into DexFuzz testing."""

  def  __init__(self, num_tests, num_inputs, device, dexer, debug_info):
    """Constructor for the tester.

    Args:
      num_tests: int, number of tests to run
      num_inputs: int, number of JFuzz programs to generate
      device: string, target device serial number (or None)
      dexer: string, defines dexer
      debug_info: boolean, if True include debugging info
    """
    self._num_tests = num_tests
    self._num_inputs = num_inputs
    self._device = device
    self._save_dir = None
    self._results_dir = None
    self._dexfuzz_dir = None
    self._inputs_dir = None
    self._dexfuzz_env = None
    self._dexer = dexer
    self._debug_info = debug_info

  def __enter__(self):
    """On entry, enters new temp directory after saving current directory.

    Raises:
      FatalError: error when temp directory cannot be constructed
    """
    self._save_dir = os.getcwd()
    self._results_dir = mkdtemp(dir='/tmp/')
    self._dexfuzz_dir = mkdtemp(dir=self._results_dir)
    self._inputs_dir = mkdtemp(dir=self._dexfuzz_dir)
    if self._results_dir is None or self._dexfuzz_dir is None or \
        self._inputs_dir is None:
      raise FatalError('Cannot obtain temp directory')
    self._dexfuzz_env = os.environ.copy()
    self._dexfuzz_env['ANDROID_DATA'] = self._dexfuzz_dir
    top = GetEnvVariableOrError('ANDROID_BUILD_TOP')
    self._dexfuzz_env['PATH'] = (top + '/art/tools/bisection_search:' +
                                 self._dexfuzz_env['PATH'])
    android_root = GetEnvVariableOrError('ANDROID_HOST_OUT')
    self._dexfuzz_env['ANDROID_ROOT'] = android_root
    self._dexfuzz_env['LD_LIBRARY_PATH'] = android_root + '/lib'
    os.chdir(self._dexfuzz_dir)
    os.mkdir('divergent_programs')
    os.mkdir('bisection_outputs')
    return self

  def __exit__(self, etype, evalue, etraceback):
    """On exit, re-enters previously saved current directory and cleans up."""
    os.chdir(self._save_dir)
    # TODO: detect divergences or shutil.rmtree(self._results_dir)

  def Run(self):
    """Feeds JFuzz programs into DexFuzz testing."""
    print()
    print('**\n**** J/DexFuzz Testing\n**')
    print()
    print('#Tests    :', self._num_tests)
    print('Device    :', self._device)
    print('Directory :', self._results_dir)
    print()
    self.GenerateJFuzzPrograms()
    self.RunDexFuzz()

  def CompileOnHost(self):
    """Compiles Test.java into classes.dex using either javac/dx or d8.

    Raises:
      FatalError: error when compilation fails
    """
    if self._dexer == 'dx' or self._dexer == 'd8':
      dbg = '-g' if self._debug_info else '-g:none'
      if RunCommand(['javac', '--release=8', dbg, 'Test.java'],
                    out=None, err='jerr.txt', timeout=30) != RetCode.SUCCESS:
        print('Unexpected error while running javac')
        raise FatalError('Unexpected error while running javac')
      cfiles = glob('*.class')
      dx = 'dx' if self._dexer == 'dx' else 'd8-compat-dx'
      if RunCommand([dx, '--dex', '--output=classes.dex'] + cfiles,
                    out=None, err='dxerr.txt', timeout=30) != RetCode.SUCCESS:
        print('Unexpected error while running dx')
        raise FatalError('Unexpected error while running dx')
      # Cleanup on success (nothing to see).
      for cfile in cfiles:
        os.unlink(cfile)
      os.unlink('jerr.txt')
      os.unlink('dxerr.txt')
    else:
      raise FatalError('Unknown dexer: ' + self._dexer)

  def GenerateJFuzzPrograms(self):
    """Generates JFuzz programs.

    Raises:
      FatalError: error when generation fails
    """
    os.chdir(self._inputs_dir)
    for i in range(1, self._num_inputs + 1):
      if RunCommand(['jfuzz'], out='Test.java', err=None) != RetCode.SUCCESS:
        print('Unexpected error while running JFuzz')
        raise FatalError('Unexpected error while running JFuzz')
      self.CompileOnHost()
      shutil.move('Test.java', '../Test' + str(i) + '.java')
      shutil.move('classes.dex', 'classes' + str(i) + '.dex')

  def RunDexFuzz(self):
    """Starts the DexFuzz testing."""
    os.chdir(self._dexfuzz_dir)
    dexfuzz_args = ['--inputs=' + self._inputs_dir,
                    '--execute',
                    '--execute-class=Test',
                    '--repeat=' + str(self._num_tests),
                    '--quiet',
                    '--dump-output',
                    '--dump-verify',
                    '--interpreter',
                    '--optimizing',
                    '--bisection-search']
    if self._device is not None:
      dexfuzz_args += ['--device=' + self._device, '--allarm']
    else:
      dexfuzz_args += ['--host']  # Assume host otherwise.
    cmd = ['dexfuzz'] + dexfuzz_args
    print('**** Running ****\n\n', cmd, '\n')
    call(cmd, env=self._dexfuzz_env)
    print('\n**** Results (report.log) ****\n')
    call(['tail', '-n 24', 'report.log'])


def main():
  # Handle arguments.
  parser = argparse.ArgumentParser()
  parser.add_argument('--num_tests', default=1000, type=int,
                      help='number of tests to run (default: 1000)')
  parser.add_argument('--num_inputs', default=10, type=int,
                      help='number of JFuzz program to generate (default: 10)')
  parser.add_argument('--device', help='target device serial number')
  parser.add_argument('--dexer', default='dx', type=str,
                      help='defines dexer as dx or d8 (default: dx)')
  parser.add_argument('--debug_info', default=False, action='store_true',
                      help='include debugging info')
  args = parser.parse_args()
  # Run the DexFuzz tester.
  with DexFuzzTester(args.num_tests,
                     args.num_inputs,
                     args.device,
                     args.dexer, args.debug_info) as fuzzer:
    fuzzer.Run()

if __name__ == '__main__':
  main()



#!/usr/bin/env python
#
# Copyright (C) 2012 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates default implementations of operator<< for enum types."""

import codecs
import re
import sys


_ENUM_START_RE = re.compile(
    r'\benum\b\s+(class\s+)?(\S+)\s+:?.*\{(\s+// private)?')
_ENUM_VALUE_RE = re.compile(r'([A-Za-z0-9_]+)(.*)')
_ENUM_END_RE = re.compile(r'^\s*\};$')
_ENUMS = {}
_NAMESPACES = {}
_ENUM_CLASSES = {}


def Confused(filename, line_number, line):
    sys.stderr.write('%s:%d: confused by:\n%s\n' %
                     (filename, line_number, line))
    raise Exception("giving up!")
    sys.exit(1)


def ProcessFile(filename):
    lines = codecs.open(filename, 'r', 'utf8', 'replace').read().split('\n')

    class EnumLines:
        def __init__(self, ns, ec):
            self.namespaces = ns
            self.enclosing_classes = ec
            self.lines = []

    def generate_enum_lines(l):
        line_number = 0
        enum_lines = None
        namespaces = []
        enclosing_classes = []

        for raw_line in l:
            line_number += 1

            if enum_lines is None:
                # Is this the start of a new enum?
                m = _ENUM_START_RE.search(raw_line)
                if m:
                    # Yes, so create new line list.
                    enum_lines = EnumLines(namespaces[:], enclosing_classes[:])
                    enum_lines.lines.append((raw_line, line_number))
                    continue

                # Is this the start or end of a namespace?
                m = re.search(r'^namespace (\S+) \{', raw_line)
                if m:
                    namespaces.append(m.group(1))
                    continue
                m = re.search(r'^\}\s+// namespace', raw_line)
                if m:
                    namespaces = namespaces[0:len(namespaces) - 1]
                    continue

                # Is this the start or end of an enclosing class or struct?
                m = re.search(
                    r'^\s*(?:class|struct)(?: MANAGED)?(?: PACKED\([0-9]\))? (\S+).* \{', raw_line)
                if m:
                    enclosing_classes.append(m.group(1))
                    continue

                # End of class/struct -- be careful not to match "do { ... } while" constructs by accident
                m = re.search(r'^\s*\}(\s+)?(while)?(.+)?;', raw_line)
                if m and not m.group(2):
                    enclosing_classes = enclosing_classes[0:len(enclosing_classes) - 1]
                    continue

                continue

            # Is this the end of the current enum?
            m = _ENUM_END_RE.search(raw_line)
            if m:
                if enum_lines is None:
                    Confused(filename, line_number, raw_line)
                yield enum_lines
                enum_lines = None
                continue

            # Append the line
            enum_lines.lines.append((raw_line, line_number))

    for enum_lines in generate_enum_lines(lines):
        m = _ENUM_START_RE.search(enum_lines.lines[0][0])
        if m.group(3) is not None:
            # Skip private enums.
            continue

        # Add an empty entry to _ENUMS for this enum.
        is_enum_class = m.group(1) is not None
        enum_name = m.group(2)
        if len(enum_lines.enclosing_classes) > 0:
            enum_name = '::'.join(enum_lines.enclosing_classes) + '::' + enum_name
        _ENUMS[enum_name] = []
        _NAMESPACES[enum_name] = '::'.join(enum_lines.namespaces)
        _ENUM_CLASSES[enum_name] = is_enum_class

        def generate_non_empty_line(lines):
            for raw_line, line_number in lines:
                # Strip // comments.
                line = re.sub(r'//.*', '', raw_line)
                # Strip whitespace.
                line = line.strip()
                # Skip blank lines.
                if len(line) == 0:
                    continue

                # The only useful thing in comments is the <<alternate text>> syntax for
                # overriding the default enum value names. Pull that out...
                enum_text = None
                m_comment = re.search(r'// <<(.*?)>>', raw_line)
                if m_comment:
                    enum_text = m_comment.group(1)

                yield (line, enum_text, raw_line, line_number)

        for line, enum_text, raw_line, line_number in generate_non_empty_line(enum_lines.lines[1:]):
            # Since we know we're in an enum type, and we're not looking at a comment
            # or a blank line, this line should be the next enum value...
            m = _ENUM_VALUE_RE.search(line)
            if not m:
                Confused(filename, line_number, raw_line)
            enum_value = m.group(1)

            # By default, we turn "kSomeValue" into "SomeValue".
            if enum_text is None:
                enum_text = enum_value
                if enum_text.startswith('k'):
                    enum_text = enum_text[1:]

            # Check that we understand the line (and hopefully do not parse incorrectly), or should
            # filter.
            rest = m.group(2).strip()

            # With "kSomeValue = kOtherValue," we take the original and skip later synonyms.
            # TODO: check that the rhs is actually an existing value.
            if rest.startswith('= k'):
                continue

            # Remove trailing comma.
            if rest.endswith(','):
                rest = rest[:-1]

            # We now expect rest to be empty, or an assignment to an "expression."
            if len(rest):
                # We want to lose the expression "= [exp]". As we do not have a real C parser, just
                # assume anything without a comma is valid.
                m_exp = re.match('= [^,]+$', rest)
                if m_exp is None:
                    sys.stderr.write('%s\n' % (rest))
                    Confused(filename, line_number, raw_line)

            # If the enum is scoped, we must prefix enum value with enum name (which is already prefixed
            # by enclosing classes).
            if is_enum_class:
                enum_value = enum_name + '::' + enum_value
            else:
                if len(enum_lines.enclosing_classes) > 0:
                    enum_value = '::'.join(enum_lines.enclosing_classes) + '::' + enum_value

            _ENUMS[enum_name].append((enum_value, enum_text))


def main():
    local_path = sys.argv[1]
    header_files = []
    for header_file in sys.argv[2:]:
        header_files.append(header_file)
        ProcessFile(header_file)

    print('#include <iostream>')
    print('')

    for header_file in header_files:
        header_file = header_file.replace(local_path + '/', '')
        print('#include "%s"' % header_file)

    print('')

    for enum_name in _ENUMS:
        print('// This was automatically generated by art/tools/generate_operator_out.py --- do not edit!')

        namespaces = _NAMESPACES[enum_name].split('::')
        for namespace in namespaces:
            print('namespace %s {' % namespace)

        print(
            'std::ostream& operator<<(std::ostream& os, const %s& rhs) {' % enum_name)
        print('  switch (rhs) {')
        for (enum_value, enum_text) in _ENUMS[enum_name]:
            print('    case %s: os << "%s"; break;' % (enum_value, enum_text))
        if not _ENUM_CLASSES[enum_name]:
            print(
                '    default: os << "%s[" << static_cast<int>(rhs) << "]"; break;' % enum_name)
        print('  }')
        print('  return os;')
        print('}')

        for namespace in reversed(namespaces):
            print('}  // namespace %s' % namespace)
        print('')

    sys.exit(0)


if __name__ == '__main__':
    main()
